(* compiler correctness proof for
   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010
*)

Require Import Arith List Program FunctionalExtensionality Heq.
Require Import util. 
Require Import ssreflect ssrbool Heq.
Require Import cesd sysf sysfopsem logrel.

Set Implicit Arguments.

Module CompilerCorrectFunctor (ObsLeMod: ObsLeSig).

Module ObsLeModInstance := ObsLeMod.
Module LogRelMod := LogicalRelationFunctor (ObsLeModInstance).
Export LogRelMod.
(*====== coqdoc printing directives ========================================*)
(** printing --> %\ensuremath{\mathrel{\texttt{->}}}% *)
(** printing ** %\ensuremath{\mathrel{\texttt{*}}}% *)

(** printing E %{\ensuremath{\Gamma}}% *)
(** printing env' %{\ensuremath{\Gamma'}}% *)
(** printing T %{\ensuremath{\tau}}% *)
(** printing ty1 %\ensuremath{\tau_1}% *)
(** printing ty2 %\ensuremath{\tau_2}% *)
(** printing ty' %\ensuremath{\tau'}% *)

(** printing tc %{\ensuremath{M}}% *)
(** printing e' %{\ensuremath{M'}}% *)
(** printing tc1 %{\ensuremath{M_1}}% *)
(** printing tc2 %{\ensuremath{M_2}}% *)
(** printing tc' %{\ensuremath{N}}% *)

(** printing tv %{\ensuremath{V}}% *)
(** printing tv1 %{\ensuremath{V_1}}% *)
(** printing tv2 %{\ensuremath{V_2}}% *)

(** printing tv' %{\ensuremath{W}}% *)

(** printing compileE %{\ensuremath{\coqdocvar{compile}_E}}% *)
(** printing compileV %{\ensuremath{\coqdocvar{compile}_V}}% *)

(** printing Ret %{\ensuremath{\mathtt{Ret}}}% *)
(** printing Push %{\ensuremath{\mathtt{Push}}}% *)
(** printing PushN %{\ensuremath{\mathtt{PushN}}}% *)
(** printing MkPair %{\ensuremath{\mathtt{MkPair}}}% *)
(** printing PushC %{\ensuremath{\mathtt{PushC}}}% *)
(** printing PushRC %{\ensuremath{\mathtt{PushRC}}}% *)
(** printing Op %{\ensuremath{\mathtt{Op}}}% *)
(** printing Sel %{\ensuremath{\mathtt{Sel}}}% *)
(** printing SelNoDump %{\ensuremath{\mathtt{SelNoDump}}}% *)
(** printing App %{\ensuremath{\mathtt{App}}}% *)
(** printing AppNoDump %{\ensuremath{\mathtt{AppNoDump}}}% *)
(** printing Join %{\ensuremath{\mathtt{Join}}}% *)
(** printing ProjFst %{\ensuremath{\mathtt{ProjFst}}}% *)
(** printing ProjSnd %{\ensuremath{\mathtt{ProjSnd}}}% *)
(** printing PushE %{\ensuremath{\mathtt{PushE}}}% *)
(** printing PopE %{\ensuremath{\mathtt{PopE}}}% *)

Definition OpN_to_OpM (op: OperationNat) : OperationM :=
  match op with
    | PLUS => M_PLUS
    | MULT => M_MULT
    | MINUS => M_MINUS
  end.

Definition OpB_to_OpM (op: OperationBool) : OperationM :=
  match op with
    | EQ => M_EQ
    | LE => M_LE
    | GT => M_GT
  end.

Definition nat_to_bool (n:nat) : bool := 
  match n with | 0 => false | _ => true end.

Lemma OpSemNat_eq: forall o n1 n2, 
  OpSemNat o n1 n2 = OpSemM (OpN_to_OpM o) n1 n2.
Proof. intros; destruct o; auto. Qed.

Lemma OpSemBool_eq: forall o n1 n2, 
  OpSemBool o n1 n2 = nat_to_bool (OpSemM (OpB_to_OpM o) n1 n2).
Proof.
intros; destruct o; simpl; 
 [ case (beq_nat n1 n2) | case (ble_nat n1 n2) | case (blt_nat n2 n1) ]; auto.
Qed.

Definition tailcode (attachret: bool) := if attachret then [Ret] else nil.

Fixpoint compileV u E T (tv : Val u E T) : Code :=
  match tv with
    | VAR _ v => [Push (idxVar v)]
    | VUNIT => [PushN 0]
    | VINT n => [PushN n]
    | PAIR _ _ tv1 tv2 => (compileV tv1) ++ (compileV tv2) ++ [MkPair]
    | INL _ _ v => [PushN 1] ++ (compileV v) ++ [MkPair]
    | INR _ _ v => [PushN 0] ++ (compileV v) ++ [MkPair]
    | LAM _ _ tc => [PushC (compileE tc true)]
    | REC _ _ tc => [PushRC (compileE tc true)]
    | TLAM _ tv' => compileV tv'
    | PACK _ ty' tv' => compileV tv'
  end

with compileE u E T (tc : Exp u E T) (attachret: bool) : Code :=
  match tc with
    | VAL _ tv => compileV tv ++ tailcode attachret
    | OPN op tv1 tv2 => (compileV tv1) ++ (compileV tv2) ++ 
      [Op (OpN_to_OpM op)] ++ tailcode attachret
    | OPB op tv1 tv2 => (compileV tv1) ++ (compileV tv2) ++ 
      [Op (OpB_to_OpM op)] ++ [PushN 0] ++ [MkPair] ++ tailcode attachret
    | FST _ _ tv => (compileV tv) ++ [ProjFst] ++ tailcode attachret
    | SND _ _ tv => (compileV tv) ++ [ProjSnd] ++ tailcode attachret
    | CASE _ _ _ tv tc1 tc2 => 
      if attachret 
      then (compileV tv) ++ [Dup; ProjSnd; PushE; ProjFst; 
        SelNoDump (compileE tc1 true) (compileE tc2 true); PopE]
      else (compileV tv) ++ [Dup; ProjSnd; PushE; ProjFst; 
        Sel (compileE tc1 false ++ [Join]) (compileE tc2 false ++ [Join]); PopE]
    | LET _ _ tc1 tc2 => 
      compileE tc1 false ++ [PushE] ++ compileE tc2 attachret ++ [PopE]
    | APP _ _ tv1 tv2 => (compileV tv1) ++ (compileV tv2) ++
      if attachret then [AppNoDump] else [App]
    | TAPP _ tv ty => compileV tv ++ tailcode attachret
    | UNPK _ ty' tv tc' => 
      (compileV tv) ++ [PushE] ++ (compileE tc' attachret) ++ [PopE]
  end.

(*
   Relations
*)

(* Le_ret *)

Definition ntop_kstepc_ret (pe:MEnv) (P: prdc iCESD) : prdc iComp := 
  fun icomp =>
  match icomp with (i,(c,s)) =>
    forall j c0 e0 s0 d0, P (j, (c0,e0,s0,d0)) ->
      forall c1 e1 s1 k, k <= i -> k <= j -> 
        stepsfor k (c++c1, pe++e1, s++s1, (c0,e0,s0)::d0)
  end.

Definition MonadLe_ret (e:MEnv) T (RT: irel MValue (CVal T)) 
  : irel MComp (CExp T) :=
  fun i comp tc => 
    ntop_kstepc_ret e (pton_kstepv nil (ivalset_le RT tc)) (i, comp).

Definition irelLeCFun_ret u srel E T 
  : irel MComp (Exp _ (subEnv (isr_sub srel) E) (subTy (isr_sub srel) T)) 
  := fun i c tc =>
    forall k e te,
      k <= i -> @irelLeEFun u srel E k e te ->
      MonadLe_ret e (@irelLeVFun u srel T) k c (SubExp te tc).
Implicit Arguments irelLeCFun_ret [u [E] [T]].

Definition irelLeC_ret u E T : irel MComp (Exp u E T) := fun i c tc =>
  forall srel, isr_decr srel -> 
    irelLeCFun_ret srel i c (subExp (isr_sub srel) tc).

Definition relLeC_ret u E T := fun comp tc =>
  exists tc', tc' <<C<< cstr tc /\ forall k, @irelLeC_ret u E T k comp tc'.

Definition ntop_termc_ret (pe:MEnv) (P: prdc CESD) : prdc MComp := fun comp =>
  match comp with (c,s) =>
    forall c0 e0 s0 d0, P (c0,e0,s0,d0) ->
      forall c1 e1 s1, terminate (c++c1, pe++e1, s++s1, (c0,e0,s0)::d0)
  end.

(* Ge_ret *)

Definition MonadGe_ret (e:MEnv) T (RT: rel MValue (CVal T)) 
  : rel MComp (CExp T) := fun comp tc => 
  forall tv, tc |==> tv ->
    ntop_termc_ret e (pton_termv nil (valset_ge RT tv)) comp.

Definition erelGeCFun_ret u srel E T 
  : rel MComp (Exp _ (subEnv (sr_sub srel) E) (subTy (sr_sub srel) T)) 
  := fun c tc =>
    forall e te,
    @erelGeEFun u srel E e te ->
    MonadGe_ret e (@erelGeVFun u srel T) c (SubExp te tc).

Definition erelGeC_ret u E T : rel MComp (Exp u E T) 
  := fun c tc =>
    forall srel,
    erelGeCFun_ret srel _ _ c (subExp (sr_sub srel) tc).

Definition relGeC_ret u E T : rel MComp (Exp u E T) 
  := fun comp tc =>
  exists tcs, tc <<C<<tcs /\ (forall p, @erelGeC_ret u E T comp (tcs p)).

(* LeCV *)

Definition irelLeCV u E T : irel MComp (Val u E T) := fun i c tv =>
  forall e cesd, length e = length E -> exists v, 
    reduce (c +~+ e +-+ cesd) (#v v +~+ e +-+ cesd) /\
    forall srel k te, 
      isr_decr srel -> k <= i -> @irelLeEFun u srel E k e te ->
      irelLeVFun srel k v (SubVal te (subVal (isr_sub srel) tv)).

Definition relLeCV u (E:Env u) T := fun c tv =>
  exists tv', tv' <<V<< (cstr tv) /\ forall k, @irelLeCV u E T k c tv'.

(* GeCV *)

Definition erelGeCV  u E T : rel MComp (Val u E T) := fun c tv =>
  forall e cesd, length e = length E -> exists v, 
    reduce (c +~+ e +-+ cesd) (#v v +~+ e +-+ cesd) /\
    forall srel te, 
      @erelGeEFun u srel E e te ->
      erelGeVFun srel v (SubVal te (subVal (sr_sub srel) tv)).

Definition relGeCV u (E:Env u) T := fun c tv =>
  exists tvs, tv <<V<< tvs /\ (forall p, @erelGeCV u E T c (tvs p)).

Notation "c '~~' i '<cv<' tv" := (irelLeCV i c tv) 
  (at level 70, no associativity).
Notation "c '>cv>' tv" := (erelGeCV c tv) (at level 70, no associativity).
Notation "c '<<cv<<' tv" := (relLeCV c tv) (at level 70, no associativity).
Notation "c '>>cv>>' tv" := (relGeCV c tv) (at level 70, no associativity).
Notation "c '<<ccret<<' tc" := (relLeC_ret c tc) (at level 70, no associativity).
Notation "c '>>ccret>>' tc" := (relGeC_ret c tc) (at level 70, no associativity).

(* cv implies cc *)

Lemma cv_implies_cc_le : forall u E T (tv: Val u E T) comp,
  comp <<cv<< tv -> comp <<cc<< ^tv.
Proof.
move => u E T tv [c v] [tv' [Htv Htv']].
exists (^tv'); split; first by apply ObsLe_VAL; Apply.
move => k srel Hsrel k0 e te Hk0k Hete j c0 e0 s0 d0 Hcont0 k1 Hk1k0 Hk1j.
red in Htv'.
specialize (Htv' k e (c0,e0,s0,d0) (irelLeEFun_len Hete)).
destruct Htv' as [v' [[k2 Hcsv'] Hv']].
eapply stepsfor_kstepreduce_rev; first by Apply.
simpl; eapply Hcont0.
eexists; split; first by unfold_SubVE; eexists; constructor.
by Apply; [ | apply Hk0k | ].
omega. omega.
Qed.

Lemma cv_implies_cc_ge : forall u E T (tv: Val u E T) comp,
  comp >>cv>> tv -> comp >>cc>> ^tv.
Proof.
move => u E T tv [c v] [tvs [Htv Htvs]].
exists (fun p => ^tvs p); split; first by apply ObsLe_VAL; Apply.
move => p srel e te Hete tv' Htv' c0 e0 s0 d0 Hcont0.
specialize (Htvs p e (c0,e0,s0,d0) (erelGeEFun_len Hete)).
destruct Htvs as [v' [[k2 Hcsv'] Hv']].
eapply terminate_kstepreduce_rev; first by Apply.
simpl; eapply Hcont0.
destruct Htv' as [i Htv']; move : Htv'.
unfold_SubVE; move => Htv'; inversion Htv'; Hsubst; clear Htv'.
by Apply.
Qed.

(*
   Correcteness Le 
*)

Lemma compiler_correctness_le : forall u E,
  (forall T (tv:Val u E T), #c (compileV tv) <<cv<< tv) /\
  (forall T (tc:Exp u E T), 
    #c (compileE tc false) <<cc<< tc /\ 
    #c (compileE tc true) <<ccret<< tc).
Proof.
apply ValExp_ind; intros; unfold CtoC.

(******* 
   VAR 
 *******)
red; intros; rename v into var. eexists. split.
- unfold cstr; apply ObsLeV_refl.
- red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
  assert (idxVar var < length e) by (rewrite H; apply idxVar_range).
  destruct (nth_error_val _ H0) as [v ?]. eexists. split; first eexists.
  - simpl; ksteps; auto.
  - intros. destruct (irelLeEFun_inv srel _ _ _ var H4) as [v' [? ?]].
    rewrite H1 in H5. depdestr H5. auto.

(******* 
   UNIT
 *******) 
red; intros. eexists. split.
- unfold cstr; apply ObsLeV_refl.
- red; intros. destruct cesd as [[[c0 e0] s0] d0]. eexists. split; first eexists.
  - simpl. ksteps. auto.
  - intros. simpl. auto.

(******* 
   INT
 *******) 
red; intros. eexists. split.
- unfold cstr; apply ObsLeV_refl.
- red; intros. destruct cesd as [[[c0 e0] s0] d0]. eexists. split; first eexists.
  - simpl. ksteps. auto.
  - intros. simpl. auto.

(******* 
   PAIR
 *******) 
red; intros. rename v into tv1, v0 into tv2.
destruct H as [tv1' [? ?]]. destruct H0 as [tv2' [? ?]]. eexists. split.
- apply (ObsLe_PAIR _ _ _ _ _ H H0).
- red; intros. destruct cesd as [[[c0 e0] s0] d0].
  withcesd c' e' s' d' (destruct (H1 k e (c',e',s',d') H3) as [v1 [[n1 ?] ?]]).
  withcesd c' e' s' d' (destruct (H2 k e (c',e',s',d') H3) as [v2 [[n2 ?] ?]]).
  eexists. split; first eexists (n1+(n2+1)). 
  - simpl in H4 |-*. rewrite ! app_ass. eapply kstepreduce_compose. apply H4.
    eapply kstepreduce_compose. simpl in H6. apply H6.
    simpl. rewrite (plus_n_O 1). ksteps. auto.
  - intros. eexists. eexists. split. 
    - unfold_SubVE. auto. 
    - split; Apply; auto with arith.

(******* 
   INL
 *******) 
red; intros. rename v into tv1.
destruct H as [tv [? ?]]. eexists. split; first by apply (ObsLe_INL _ ty2 _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0].
withcesd c' e' s' d' (destruct (H0 k e (c',e',s',d') H1) as [v1 [[n1 ?] ?]]).
eexists. split; first eexists.
- simpl. ksteps. simpl in H2 |- *. rewrite ! app_ass.
  eapply kstepreduce_compose. apply H2. ksteps. auto.
- intros. eexists. split. unfold_SubVE. reflexivity. auto.

(******* 
   INR
 *******) 
red; intros. rename v into tv1.
destruct H as [tv [? ?]]. eexists. split; first by apply (ObsLe_INR _ ty1 _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0].
withcesd c' e' s' d' (destruct (H0 k e (c',e',s',d') H1) as [v1 [[n1 ?] ?]]).
eexists. split; first eexists.
- simpl. ksteps. simpl in H2 |- *. rewrite ! app_ass.
  eapply kstepreduce_compose. apply H2. ksteps. auto.
- intros. eexists. split. unfold_SubVE. reflexivity. auto.

(******* 
   LAM
 *******)
red; intros. rename e into tc.
destruct H. clear H. destruct H0 as [tc' [? ?]]. 
eexists. split; first by apply (ObsLe_LAM _ _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
eexists. split; first by eexists; simpl; ksteps.
do 4 red. unfold_SubVE. fold irelLeVFun. intros.
simpl. ssteps. rewrite -> app_comm_cons, (app_nil_end (compileE tc true)).
eapply H0. 
- apply H2. 
- apply (le_n k2).
- instantiate (1:= consMap tv te). split.
  - rewrite hdConsMap. apply irelLeVFun_decr with k1; [ auto | omega | auto ].
  - rewrite tlConsMap. apply irelLeEFun_decr with k0; [ auto | omega | auto ].
- red; intros. eapply H7. 
  - destruct H10 as [tv' [[n ?] ?]]. eexists. split.
    - eexists. eapply Ev_step. constructor. simpl_Sub. apply H10.
    - apply H13.
- omega. - apply H12. - omega. - omega.

(******* 
   REC
 *******) 
red; intros. rename e into tc.
destruct H. clear H. destruct H0 as [tc' [? ?]]. 
eexists; split; first by apply (ObsLe_REC _ _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
eexists. split; first by eexists; simpl; ksteps.
intros; unfold_SubVE. 
cut (forall i, i <= k0 ->
  irelLeVFun srel i (VRC (e ++ e0) (compileE tc true)) 
  (SubVal te (subVal (isr_sub srel) (REC tc'))));
 first by intro; apply H5; omega.
induction i; intros; first by apply irelLeVFun_arr_zero.
do 4 red. fold irelLeVFun in *. intros.
simpl. eapply stepsfor_step_strict. constructor.
rewrite -> ! app_comm_cons, (app_nil_end (compileE tc true)). eapply H0. 
- by apply H2. 
- by apply (le_n (Peano.pred k2)).
- instantiate (1:= consMap tv (consMap 
  (REC (SubExp (SubL _ (SubL _ te)) (subExp (isr_sub srel) tc'))) te)).
  split; first by rewrite hdConsMap; apply irelLeVFun_decr with k1; [ |omega| ].
  split. 
  - rewrite -> tlConsMap, hdConsMap; apply irelLeVFun_decr with i; 
     [ auto | omega | ].
    by unfold_SubVE; rewrite <- app_nil_end; apply IHi; omega.
  - by rewrite -> tlConsMap, tlConsMap; apply irelLeEFun_decr with k0; 
     [ |omega| ].
- red; intros. eapply H8. 
  - instantiate (1:= k3). destruct H11 as [tv' [[k4 ?] ?]]. eexists. split. 
    - eexists. simpl. fold_Sub. eapply Ev_step; first by constructor. 
      by simpl_Sub; apply H11.
    - by eapply irelLeVFun_decr with i0; [ |omega| ].
  - omega. - apply H13. 
- omega. - omega.

(******* 
   TLAM
 *******) 
red; intros. rename v into tv.
destruct H as [tv' [? ?]]. eexists. split; first by apply (ObsLe_TLAM _ _ _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0].
assert (length e = length (subEnv shsub env)) by (rewrite subEnv_length; auto).
destruct (H0 k e (c0,e0,s0,d0) H2) as [v [[n ?] ?]]. 
eexists. split; first by eexists; apply H3.
intros. unfold_SubVE. red; intros.
assert (C0 : (subEnv (isr_sub srel) env, nil:Env 0) = 
  (subEnv (isr_sub (isr_cons t tr srel)) (subEnv shsub env), nil))
 by (simpl_sub).
eapply MonadLe_val.
- apply H4; [ by apply isr_cons_decr | by [] | ].
  Hhide_evars. Hrefl (hev0 : SSub PSub). 
  Hrewritec (irelLeEFun_sub env shsub (isr_cons t tr srel)). Hshow_evars.
  instantiate (1:= << SSub,PSub # C0 >> te). 
  rewrite subiSR_shsub_cons. Hsimplc. auto.
- Hrefl (SubVal (<< SSub,PSub # C0 >> te) 
  (subVal (isr_sub (isr_cons t tr srel)) tv') : SVal PVal); Helimc C1.
  intros. eapply eq_ind. eexists. apply e_TApp.
  apply (Heq_eq SVal PVal). rewrite subVal_SubVal. Hsimpl.
  Hrewrite (subSub_ss [|t|] shsub te). rewrite consmap_shsub. 
  Hrewrite (subSub_idsub te). Hsimpl.
  Helimc C2. Hrewritec (subVal_ss tv' [|t|] (subL (isr_sub srel))). 
  simpl_sub; rewrite isr_sub_cons; Hunify. auto.

(******* 
   PACK
 *******) 
red; intros. rename v into tv.
destruct H as [tv' [? ?]]. 
eexists. split; first by apply (ObsLe_PACK _ _ _ _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
destruct (H0 k e (c0,e0,s0,d0) H1) as [v [[n ?] ?]]. 
eexists. split; first by eexists; apply H2.
intros. unfold_SubVE.
exists (subTy (isr_sub srel) ty'). exists (@irelLeVFun u srel ty'). eexists. 
split; first by apply irelLeVFun_decr. 
split; first by reflexivity.
Hsimpl. rewrite <- subiSR_singsub_isrel. intros.
Hrewrite <- (irelLeVFun_sub ty [|ty'|] srel). Helimc. Helimc. apply H3; auto.

(******* 
   VAL
 *******) 
split; 
(
red; intros; rename v into tv;
destruct H as [tv' [? ?]]; eexists; split; first (by apply ObsLe_VAL; Apply);
do 4 red; intros; simpl; try rewrite <- app_nil_end; try rewrite app_ass;
(withcesd c' e' s' d' 
 (destruct (H0 k e (c',e',s',d') (irelLeEFun_len H3)) as [v [[n ?] ?]]));
eapply stepsfor_kstepreduce_rev; first (by simpl in H7; apply H7);
simpl; ssteps; eapply H4; 
[ eexists; split; 
  [ by unfold_SubVE; eexists; apply e_Val 
  | by apply H8; [ | instantiate (1:= k0); omega | ] ] 
| by omega | by omega ]
).

(******* 
   OPN
 *******) 
split;
(
red; intros; rename v into tv1, v0 into tv2;
destruct H as [tv1' [? ?]]; destruct H0 as [tv2' [? ?]];
eexists; split; first (by apply (ObsLe_OPN _ o _ _ _ _ H H0));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d' 
 (destruct (H1 k e (c',e',s',d') (irelLeEFun_len H5)) as [v1 [[n1 ?] ?]]));
(withcesd c' e' s' d'
 (destruct (H2 k e (c',e',s',d') (irelLeEFun_len H5)) as [v2 [[n2 ?] ?]]));
eapply stepsfor_kstepreduce_rev; first (by simpl in H9; apply H9);
simpl; assert (Htv1':= H10 _ k0 _ H3 H4 H5);
eapply stepsfor_kstepreduce_rev; first (by simpl in H11; apply H11); 
simpl; assert (Htv2':= H12 _ k0 _ H3 H4 H5);
destruct v1; try contradiction Htv1'; try contradiction Htv2';
destruct v2; try contradiction Htv1'; try contradiction Htv2';
ssteps; eapply H6;
[ eexists; split;
  [ unfold_SubVE; eexists; rewrite -> Htv1', Htv2'; apply e_OpN
  | simpl; rewrite OpSemNat_eq; auto ]
| apply le_n | omega ]
).

(******* 
   OPB
 *******) 
split;
(
red; intros; rename v into tv1, v0 into tv2;
destruct H as [tv1' [? ?]]; destruct H0 as [tv2' [? ?]]; 
eexists; split; first (by apply (ObsLe_OPB _ o _ _ _ _ H H0));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H1 k e (c',e',s',d') (irelLeEFun_len H5)) as [v1 [[n1 ?] ?]]));
(withcesd c' e' s' d'
 (destruct (H2 k e (c',e',s',d') (irelLeEFun_len H5)) as [v2 [[n2 ?] ?]]));
eapply stepsfor_kstepreduce_rev; first (by simpl in H9; apply H9);
simpl; assert (Htv1':= H10 _ k0 _ H3 H4 H5);
eapply stepsfor_kstepreduce_rev; first (by simpl in H11; apply H11); 
simpl; assert (Htv2':= H12 _ k0 _ H3 H4 H5);
destruct v1; try contradiction Htv1'; try contradiction Htv2';
destruct v2; try contradiction Htv1'; try contradiction Htv2';
ssteps; eapply H6;
[ eexists; split;
  [ unfold_SubVE; eexists; rewrite -> Htv1', Htv2'; apply e_OpB 
  | simpl; rewrite OpSemBool_eq;
    case (OpSemM (OpB_to_OpM o) n n0); intros; simpl;
    eexists; split; reflexivity ]
| apply le_n | omega ]
).

(******* 
   FST
 *******)
split;
(
red; intros; rename v into tv;
destruct H as [tv' [? ?]]; eexists; split; first (by apply (ObsLe_FST _ _ _ H));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H0 k e (c',e',s',d') (irelLeEFun_len H3)) as [v [[n ?] ?]]));
eapply stepsfor_kstepreduce_rev; first (by simpl in H7; apply H7); 
simpl; assert (Htv':= H8 _ k0 _ H1 H2 H3);
destruct v; try contradiction Htv';
ssteps; eapply H4;
[ destruct Htv' as [tv1 [tv2 [? [? ?]]]]; fold_Sub; fold irelLeVFun in *; 
  eexists; split;
  [ unfold_SubVE; eexists; rewrite H9; apply e_Fst
  | apply H10 ]
| omega | omega ]
).

(******* 
   SND
 *******)
split;
(
red; intros; rename v into tv;
destruct H as [tv' [? ?]]; eexists; split; first (by apply (ObsLe_SND _ _ _ H));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H0 k e (c',e',s',d') (irelLeEFun_len H3)) as [v [[n ?] ?]]));
eapply stepsfor_kstepreduce_rev; first (by simpl in H7; apply H7); 
simpl; assert (Htv':= H8 _ k0 _ H1 H2 H3);
destruct v; try contradiction Htv';
ssteps; eapply H4;
[ destruct Htv' as [tv1 [tv2 [? [? ?]]]]; fold_Sub; fold irelLeVFun in *; 
  eexists; split;
  [ unfold_SubVE; eexists; rewrite H9; apply e_Snd
  | apply H11 ]
| omega | omega ]
).

(******* 
   CASE
 *******)
destruct_pairs; split;
red; intros; rename v into tv, e into tc1, e0 into tc2; 
destruct H as [tv' [? ?]];
[ destruct H0 as [tc1' [? ?]]; destruct H1 as [tc2' [? ?]];
  eexists; split; first (by apply (ObsLe_CASE _ _ _ _ _ _ _ H H0 H1))
| destruct H3 as [tc1' [? ?]]; destruct H2 as [tc2' [? ?]];
  eexists; split; first (by apply (ObsLe_CASE _ _ _ _ _ _ _ H H3 H2)) ];
(
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H4 k e (c',e',s',d') (irelLeEFun_len H9)) as [v [[n ?] ?]]));
eapply stepsfor_kstepreduce_rev; first (by simpl in H13; apply H13);
assert (Htv':= H14 _ k0 _ H7 H8 H9); 
destruct v; try contradiction Htv'; destruct v1; try contradiction Htv';
ssteps; destruct n0;
( destruct Htv' as [tv'' [? ?]]; fold_Sub; fold irelLeVFun in *;
  ssteps; rewrite app_comm_cons; try rewrite (app_nil_end (compileE _ true));
  EApply;
  [ by apply H7
  | by apply (le_n k1)
  | instantiate (1:= consMap tv'' te); split;
    [ by rewrite hdConsMap; apply irelLeVFun_decr with k0; [ | omega | ]
    | by rewrite tlConsMap; apply irelLeEFun_decr with k0; [ | omega | ] ]
  | red; intros; ssteps; eapply H10;
    [ unfold_SubVE; rewrite H15; destruct H17 as [tv''' [[? ?] ?]]; 
      eexists; split;
      [ eexists; constructor; simpl_Sub; apply H17
      | apply H20 ]
    | omega | apply H19 ]
  | omega | omega ] )
).

(******* 
   LET
 *******)
destruct_pairs; split;
red; intros; rename e into tc1, e0 into tc2;
[ destruct H as [tc1' [? ?]]; destruct H0 as [tc2' [? ?]];
  eexists; split; first (by apply (ObsLe_LET _ _ _ _ _ H H0))
| destruct H as [tc1' [? ?]]; destruct H1 as [tc2' [? ?]];
   eexists; split; first (by apply (ObsLe_LET _ _ _ _ _ H H1)) ];
(
do 4 red; intros; simpl; rewrite ! app_ass; eapply H3;
[ apply H5
| apply (le_n k1)
| apply irelLeEFun_decr with k0; [ auto | omega | apply H7 ]
| instantiate (1:= k1); red; intros; rewrite <- app_comm_cons;
  ssteps; rewrite app_ass;
  destruct H11 as [tvv [[stp ?] ?]]; rewrite app_comm_cons; eapply H4;
  [ apply H5
  | apply (le_n k2)
  | instantiate (1:= consMap tvv te); split;
    [ rewrite hdConsMap; apply irelLeVFun_decr with i; [ auto | omega | auto ]
    | rewrite tlConsMap; apply irelLeEFun_decr with k0; [ auto | omega | auto ] ]
  | instantiate (1:= k2); red; intros; ssteps; simpl;
    destruct H15 as [tvv' [[stp' ?] ?]]; eapply H8;
    [ eexists; split;
      [ unfold_SubVE; eexists; eapply e_Let; first (by apply H11); 
        simpl_Sub; apply H15
      | apply H18 ]
    | omega | omega ]
  | omega | omega ]
| omega | omega ]
).

(******* 
   APP
 *******)
destruct_pairs; split;
(
red; intros; rename v into tf, v0 into tv;
destruct H as [tf' [? ?]]; destruct H0 as [tv' [? ?]];
eexists; split; first (by apply (ObsLe_APP _ _ _ _ _ H H0));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H1 k1 e (c',e',s',d') (irelLeEFun_len H5)) as [f [[n ?] ?]]));
(withcesd c' e' s' d'
 (destruct (H2 k1 e (c',e',s',d') (irelLeEFun_len H5)) as [v [[n' ?] ?]]));
eapply stepsfor_kstepreduce_rev; first (by simpl in H9; apply H9);
eapply stepsfor_kstepreduce_rev; first (by simpl in H11; apply H11)
).
(* attachret = false *)
change (v::f::s) with ([v;f]++s); eapply H10.
- apply H3. 
- apply le_n.
- apply irelLeEFun_decr with k0; [ auto | omega | apply H5 ].
- apply le_n.
- apply H12; auto; try omega.
  apply irelLeEFun_decr with k0; [ auto | omega | apply H5 ].
- apply H6.
- omega. - omega.
(* attachret = true *)
apply stepsfor_decr with k1; first (by omega).
assert (stepsfor k1 ([App]++c0, e0, [v;f]++s0, d0)). 
- eapply H10. 
  - apply H3. 
  - apply le_n.
  - apply irelLeEFun_decr with k0; [ auto | omega | apply H5 ].
  - apply le_n.
  - apply H12; auto; try omega.
    apply irelLeEFun_decr with k0; [ auto | omega | apply H5 ].
  - apply H6.
  - omega. - omega.
- destruct k1; first (by apply stepsfor_zero).
  destruct H13 as [k' [cesd' [? ?]]].
  destruct k'; first (by inversion H13).
  depdestr H14. simpl in H14. depdestr H14.
  - eapply stepsfor_step_strict. constructor. 
    exists k'. eexists. split; [ omega | apply H15 ].
  - eapply stepsfor_step_strict. constructor. 
    exists k'. eexists. split; [ omega | apply H15 ].

(******* 
   TAPP
 *******)
split;
(
red; intros; rename v into tf; destruct H as [tf' [? ?]]; 
eexists; split; first (by apply (ObsLe_TAPP _ _ _ ty' H));
do 4 red; intros;
(withcesd c' e' s' d'
 (destruct (H0 k1 e (c',e',s',d') (irelLeEFun_len H3)) as [fv [[n ?] ?]]));
simpl in H7 |- *; try rewrite <- app_nil_end; try rewrite app_ass;
eapply stepsfor_kstepreduce_rev; first (by apply H7); 
apply stepsfor_decr with k1; first (by omega);
ssteps; rewrite (cons_app fv); eapply H8;
[ apply H1 
| apply le_n
| apply irelLeEFun_decr with k0; [ auto | omega | apply H3 ]
| apply irelLeVFun_decr; apply H1
| fold irelLeVFun (app e e0); instantiate (1:= k1); red; intros; simpl;
  instantiate (1:= ty') in H9; destruct H9 as [tfx [[j' ?] ?]]; eapply H4;
  [ assert (C0 : (nil:Env 0, subTy (isr_sub (isr_cons 
    (subTy (isr_sub srel) ty') (irelLeVFun srel) srel)) ty) =
    (nil:Env 0, subTy (isr_sub srel) (subTy [|ty'|] ty)))
     by (rewrite -> isr_sub_cons, subTy_ss, isr_singsub; auto);
    eexists; split; 
    [ unfold_SubVE; Hhide_evars; Hrefl (hev0 : SVal PVal); 
      Helimc C; unfold_SubVE; Hshow_evars;
      instantiate (1 := << SVal,PVal # C0 >> tfx); Hsimpl; intros;
      Hhide_concl; revert H9; Hrefl (tfx : SVal PVal); Helimc C; Hshow_concl;
      Hunify; intros; eexists; apply H9 
    | instantiate (2:=i); Hrewrite (irelLeVFun_sub ty [|ty'|] srel); Helimc C0;
      rewrite subiSR_singsub_isrel; Hclear; auto ]
  | omega | omega ]
| omega | omega ]
). 

(******* 
   UNPK
 *******)
destruct_pairs; split;
red; intros; rename v into tv, e into tc; destruct H as [tv' [? ?]];
[ destruct H0 as [tc' [? ?]]; 
  eexists; split; first (by apply (ObsLe_UNPK _ _ _ _ _ _ H H0))
| destruct H1 as [tc' [? ?]]; 
  eexists; split; first (by apply (ObsLe_UNPK _ _ _ _ _ _ H H1))];
(
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H2 k e (c',e',s',d') (irelLeEFun_len H6)) as [v [[n ?] ?]]));
destruct (H11 srel k0 te H4 H5 H6) as [t' [tr [tvv [? [? ?]]]]];
fold_Sub; fold irelLeVFun in *;
eapply stepsfor_kstepreduce_rev; simpl; first (by simpl in H10; apply H10);
ssteps; apply stepsfor_decr with k1; first (by omega);
rewrite -> app_ass, app_comm_cons;
assert (C0 : (subEnv (isr_sub srel) env, nil:Env 0) = 
  (subEnv (isr_sub (isr_cons t' tr srel)) (subEnv shsub env), nil))
 by (rewrite -> subEnv_ss, isr_sub_cons, consmap_shsub; auto);
eapply H3;
[ instantiate (1:= isr_cons _ tr srel); apply isr_cons_decr; auto
| apply (le_n k1)
| instantiate (1:= consMap _ _); split;
  [ by rewrite hdConsMap; apply irelLeVFun_decr with k0;
    [ apply isr_cons_decr | omega | apply H14 ]
  | rewrite tlConsMap; apply irelLeEFun_decr with k0;
    [ by apply isr_cons_decr | by omega | ]; 
    Hhide_evars; Hrefl (hev0 : SSub PSub); fold_Sub; 
    Hrewritec (irelLeEFun_sub env shsub (isr_cons t' tr srel)); Hshow_evars;
    rewrite (subiSR_shsub_cons tr srel); 
    instantiate (1:= << SSub,PSub # C0 >> te); Hsimplc; auto ]
| red; intros; simpl; ssteps; destruct H15 as [tvv' [[j' ?] ?]];
  assert (C1 : (nil:Env 0, subTy (isr_sub (isr_cons t' tr srel)) 
    (subTy shsub ty2)) = (nil, subTy (isr_sub srel) ty2))
   by (rewrite -> isr_sub_cons, subTy_ss, consmap_shsub; auto);
  eapply H7;
  [ eexists; split;
    [ unfold_SubVE; eexists; rewrite H13; apply e_Unpk;
      simpl_Sub; fold_Sub; simpl_Sub;
      instantiate (1:= << SVal,PVal # C1 >> tvv'); instantiate (1:= j');
      Hrewrite (subSub_ss [|t'|] shsub te); rewrite consmap_shsub;
      Hrewrite (subSub_idsub te); Hsimpl;
      Helimc C4; Hrewrite (subExp_ss tc' [|t'|] (subL (isr_sub srel))); 
      rewrite -> consmap_subL, idsub_ss, <- (isr_sub_cons tr srel);
      Hrefl (tvv : SVal PVal); Helimc C1;
      intros; revert H15; Hunify; auto
    | instantiate (2:= i); 
      Hrefl (irelLeVFun srel (T:=ty2) : SirelV _); Helimc C1; 
      intros; revert H18;
      Hrewrite (irelLeVFun_sub ty2 shsub (isr_cons t' tr srel));
      rewrite subiSR_shsub_cons; Hunify; auto ]
  | omega | apply H17 ]
| omega | omega ]
).

Qed.

(* 
   Correcteness Ge 
*)

Lemma compiler_correctness_ge : forall u E,
  (forall T (tv:Val u E T), #c (compileV tv) >>cv>> tv) /\
  (forall T (tc:Exp u E T), 
    #c (compileE tc false) >>cc>> tc /\ 
    #c (compileE tc true) >>ccret>> tc).
Proof.
apply ValExp_ind; intros; unfold CtoC.

(******* 
   VAR 
 *******)
red; intros; rename v into var.  
exists (fun _:nat => VAR var); split; first by apply ObsLe_VAR.
red; intros; simpl; destruct cesd as [[[c0 e0] s0] d0].
assert (idxVar var < length e) by (rewrite H; apply idxVar_range).
destruct (nth_error_val _ H0) as [v ?]. 
eexists; split; first by eexists; simpl; ksteps.
intros; destruct (erelGeEFun_inv srel _ _ var H2) as [v' [? ?]].
rewrite H1 in H3. depdestr H3. auto.

(******* 
   UNIT
 *******) 
red; intros; eexists (fun _:nat => VUNIT). split; first by apply ObsLe_UNIT. 
intros. simpl. red; intros. destruct cesd as [[[c0 e0] s0] d0].
eexists. split; first by eexists; simpl; ksteps. 
unfold_SubVE. auto.

(******* 
   INT
 *******) 
red; intros; eexists (fun _:nat => VINT n). split; first by apply ObsLe_INT. 
intros. simpl. red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
eexists. split; first by eexists; simpl; ksteps.
unfold_SubVE. auto.

(******* 
   PAIR
 *******) 
red; intros. rename v into tv1, v0 into tv2.
destruct H as [tvs1 [? ?]]. destruct H0 as [tvs2 [? ?]]. 
exists (fun i => PAIR (tvs1 i) (tvs2 i)). 
split; first by apply (ObsLe_PAIR _ _ _ _ _ H H0).
red; intros. destruct cesd as [[[c0 e0] s0] d0].
withcesd c' e' s' d' (destruct (H1 p e (c',e',s',d') H3) as [v1 [[n1 ?] ?]]).
withcesd c' e' s' d' (destruct (H2 p e (c',e',s',d') H3) as [v2 [[n2 ?] ?]]).
eexists; split; first exists (n1+(n2+1)).
- simpl; rewrite ! app_ass. 
  eapply kstepreduce_compose. simpl in H4. apply H4.
  eapply kstepreduce_compose. simpl in H6. apply H6.
  simpl. rewrite (plus_n_O 1). by ksteps.
- intros. eexists. eexists. split. unfold_SubVE. reflexivity. auto.

(******* 
   INL
 *******) 
red; intros; rename v into tv1.
destruct H as [tvs [? ?]]. exists (fun i => INL ty2 (tvs i)).
split; first by apply (ObsLe_INL _ ty2 _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0].
withcesd c' e' s' d' (destruct (H0 p e (c',e',s',d') H1) as [v1 [[n1 ?] ?]]).
eexists. split; first eexists.
- simpl; ksteps; simpl in H2 |- *. 
  rewrite ! app_ass. eapply kstepreduce_compose; by [apply H2 | ksteps].
- intros. eexists. split. unfold_SubVE. reflexivity. auto.

(******* 
   INR
 *******) 
red; intros. rename v into tv2.
destruct H as [tvs [? ?]]. exists (fun i => INR ty1 (tvs i)). 
split; first by apply (ObsLe_INR _ ty1 _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0].
withcesd c' e' s' d' (destruct (H0 p e (c',e',s',d') H1) as [v1 [[n1 ?] ?]]).
eexists. split; first eexists.
- simpl; ksteps; simpl in H2 |- *. 
  rewrite ! app_ass. eapply kstepreduce_compose; by [apply H2 | ksteps].
- intros. eexists. split. unfold_SubVE. reflexivity. auto.

(******* 
   LAM
 *******) 
red; intros. rename e into tc.
destruct H. clear H. destruct H0 as [tcs [? ?]]. 
exists (fun i => LAM (tcs i)). split; first by apply (ObsLe_LAM _ _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0]. eexists.
split; first by eexists; simpl; ksteps.
intros. unfold_SubVE. do 2 red. intros. simpl. tsteps.
rewrite -> app_comm_cons, (app_nil_end (compileE tc true)). eapply H0. 
- instantiate (1:= consMap tv te). 
  split; by [rewrite hdConsMap | rewrite tlConsMap].
- destruct H4. destruct x; first by inversion H4. eexists. eapply Ev_step_rev. 
  - apply H4.
  - eapply eq_ind; [constructor | simpl_Sub].
- auto.

(******* 
   REC
 *******) 
red; intros. rename e into tc.
destruct H. clear H. destruct H0 as [tcs [? ?]]. 
exists (fun i => RECN i (tcs i)). split; first by apply REC_approximation.
red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
eexists. split; first by eexists; simpl; ksteps. 
intros. pattern p at 1. apply generalize_induction. 
induction k; intros; first by apply BotF_bot.
do 4 red. fold erelGeVFun. intros. simpl. 
destruct H5 as [j' ?]. destruct j'; first by inversion H5.
tsteps. rewrite -> ! app_comm_cons, (app_nil_end (compileE tc true)). 
eapply (H0 p srel). 
- instantiate (1:= consMap tv (consMap (SubVal te (subVal (sr_sub srel) 
   (RECN k (tcs p)))) te)).
  split; first by rewrite hdConsMap.
  split; last by rewrite -> tlConsMap, tlConsMap.
  rewrite -> tlConsMap, hdConsMap, <- app_nil_end; apply IHk; omega.
- eexists. eapply Ev_step_rev; first by apply H5.
  fold_Sub. rewrite -> subRECN, SubRECN. fold_Sub.
  eapply eq_ind; first by constructor.
  simpl_Sub. by rewrite -> subRECN, SubRECN.
- red; intros. by apply H6.

(******* 
   TLAM
 *******) 
red; intros. rename v into tv. destruct H as [tvs [? ?]]. 
exists (fun i => TLAM (tvs i)). split; first by apply (ObsLe_TLAM _ _ _ _ H).
red; intros. 
destruct (H0 p e cesd) as [v [[n ?] ?]]; first by rewrite subEnv_length.
exists v. split; first by exists n.
intros. unfold_SubVE. intros.
assert (C0 : (subEnv (sr_sub srel) env, nil:Env 0) = 
  (subEnv (sr_sub (sr_cons t tr srel)) (subEnv shsub env), nil))
 by (simpl_sub).
eapply MonadGe_val. 
- apply H3. Hhide_evars. Hrefl (hev0 : SSub PSub). 
  Hrewritec (erelGeEFun_sub env shsub (sr_cons t tr srel)). Hshow_evars.
  instantiate (1:= << SSub,PSub # C0 >> te). 
  rewrite subSR_shsub_cons. Hsimplc. auto.
- Hrefl (SubVal (<< SSub,PSub # C0 >> te) 
   (subVal (sr_sub (sr_cons t tr srel)) (tvs p)) : SVal PVal). 
  Helimc C1.
  intros. eapply eq_ind; first by (eexists; apply e_TApp).
  apply (Heq_eq SVal PVal). rewrite subVal_SubVal. Hsimpl.
  Hrewrite (subSub_ss [|t|] shsub te). rewrite consmap_shsub. 
  Hrewrite (subSub_idsub te). Hsimpl.
  Helimc C2. Hrewritec (subVal_ss (tvs p) [|t|] (subL (sr_sub srel))).
  simpl_sub; rewrite -> sr_sub_cons; Hunify; auto.

(******* 
   PACK
 *******) 
red; intros. rename v into tv.
destruct H as [tvs [? ?]]. 
exists (fun i => PACK (tvs i)). split; first by apply (ObsLe_PACK _ _ _ _ _ H).
red; intros. destruct cesd as [[[c0 e0] s0] d0]. 
destruct (H0 p e (c0,e0,s0,d0) H1) as [v [[n ?] ?]].
eexists. split; first by eexists; apply H2.
intros. unfold_SubVE. 
exists (subTy (sr_sub srel) ty'). exists (@erelGeVFun u srel ty'). eexists. 
split; first by reflexivity.
Hsimpl. rewrite <- subSR_singsub_srel. intros.
Hrewrite <- (erelGeVFun_sub ty [|ty'|] srel). Helimc. Helimc.
apply H3; auto.

(******* 
   VAL
 *******) 
split;
(
red; intros; rename v into tv; destruct H as [tvs [? ?]];
eexists (fun i => VAL (tvs i)); split; first (by apply ObsLe_VAL; Apply);
do 4 red; intros; simpl; try rewrite <- app_nil_end; try rewrite app_ass;
(withcesd c' e' s' d' 
 (destruct (H0 p e (c',e',s',d') (erelGeEFun_len H1)) as [v [[n ?] ?]]));
eapply terminate_kstepreduce_rev; first (by simpl in H4; apply H4); 
simpl; tsteps; eapply H3;
intros; rewrite (Evv_VAL_inv H2); apply H5; auto
).

(******* 
   OPN
 *******) 
split;
(
red; intros; rename v into tv1, v0 into tv2;
destruct H as [tvs1 [? ?]]; destruct H0 as [tvs2 [? ?]];
exists (fun i => OPN o (tvs1 i) (tvs2 i)); 
split; first (by apply (ObsLe_OPN _ o _ _ _ _ H H0));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H1 p e (c',e',s',d') (erelGeEFun_len H3)) as [v1 [[n1 ?] ?]]));
(withcesd c' e' s' d'
 (destruct (H2 p e (c',e',s',d') (erelGeEFun_len H3)) as [v2 [[n2 ?] ?]]));
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H6; apply H6);
assert (Htv1':= H7 _ _ H3);
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H8; apply H8);
assert (Htv2':= H9 _ _ H3);
destruct v1; try contradiction Htv1'; try contradiction Htv2';
destruct v2; try contradiction Htv1'; try contradiction Htv2';
tsteps; eapply H5;
do 2 red; simpl; revert H4; unfold_SubVE; intro; destruct H4;
simpl in Htv1', Htv2'; rewrite -> Htv1', Htv2' in H4;
rewrite (proj2 (Ev_uniq H4 (e_OpN _ _ _))); f_equal; apply OpSemNat_eq
).

(******* 
   OPB
 *******) 
split;
(
red; intros; rename v into tv1, v0 into tv2;
destruct H as [tvs1 [? ?]]; destruct H0 as [tvs2 [? ?]];
exists (fun i => OPB o (tvs1 i) (tvs2 i)); 
split; first (by apply (ObsLe_OPB _ o _ _ _ _ H H0));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H1 p e (c',e',s',d') (erelGeEFun_len H3)) as [v1 [[n1 ?] ?]]));
(withcesd c' e' s' d'
 (destruct (H2 p e (c',e',s',d') (erelGeEFun_len H3)) as [v2 [[n2 ?] ?]]));
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H6; apply H6);
assert (Htv1':= H7 _ _ H3);
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H8; apply H8);
assert (Htv2':= H9 _ _ H3);
destruct v1; try contradiction Htv1'; try contradiction Htv2';
destruct v2; try contradiction Htv1'; try contradiction Htv2';
tsteps; eapply H5;
do 2 red; simpl; revert H4; unfold_SubVE; intro; destruct H4;
simpl in Htv1', Htv2'; rewrite -> Htv1', Htv2' in H4;
rewrite (proj2 (Ev_uniq H4 (e_OpB _ _ _))); rewrite OpSemBool_eq;
case (OpSemM (OpB_to_OpM o) n n0); eexists; split; simpl; reflexivity 
).

(******* 
   FST
 *******)
split;
(
red; intros; rename v into tv; destruct H as [tvs [? ?]]; 
exists (fun i => FST (tvs i)); split; first (by apply (ObsLe_FST _ _ _ H));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H0 p e (c',e',s',d') (erelGeEFun_len H1)) as [v [[n ?] ?]]));
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H4; apply H4);
assert (Htv':= H5 _ _ H1);
destruct v; try contradiction Htv';
simpl; tsteps; eapply H3;
destruct Htv' as [tv1 [tv2 [? [? ?]]]]; fold_Sub; fold erelGeVFun in *;
revert H2; unfold_SubVE; intro; rewrite H6 in H2; destruct H2; 
rewrite (proj2 (Ev_uniq H2 (e_Fst _ _))); auto
).

(******* 
   SND
 *******)
split;
(
red; intros; rename v into tv; destruct H as [tvs [? ?]]; 
exists (fun i => SND (tvs i)); split; first (by apply (ObsLe_SND _ _ _ H));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H0 p e (c',e',s',d') (erelGeEFun_len H1)) as [v [[n ?] ?]]));
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H4; apply H4);
assert (Htv':= H5 _ _ H1);
destruct v; try contradiction Htv';
simpl; tsteps; eapply H3;
destruct Htv' as [tv1 [tv2 [? [? ?]]]]; fold_Sub; fold erelGeVFun in *;
revert H2; unfold_SubVE; intro; rewrite H6 in H2; destruct H2; 
rewrite (proj2 (Ev_uniq H2 (e_Snd _ _))); auto
).

(******* 
   CASE
 *******)
destruct_pairs; split;
red; intros; rename v into tv, e into tc1, e0 into tc2;
destruct H as [tvs [? ?]];
[ destruct H0 as [tcs1 [? ?]]; destruct H1 as [tcs2 [? ?]]
| destruct H3 as [tcs1 [? ?]]; destruct H2 as [tcs2 [? ?]] ]; 
exists (fun i => CASE (tvs i) (tcs1 i) (tcs2 i)); 
[ split; first (by apply (ObsLe_CASE _ _ _ _ _ _ _ H H0 H1)) 
| split; first (by apply (ObsLe_CASE _ _ _ _ _ _ _ H H3 H2)) ];
(
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H4 p e (c',e',s',d') (erelGeEFun_len H7)) as [v [[n ?] ?]]));
eapply terminate_kstepreduce_rev; first (by simpl in H10; apply H10);
assert (Htv':= H11 _ _ H7);
destruct v; try contradiction Htv'; destruct v1; try contradiction Htv';
tsteps; destruct n0; 
( destruct Htv' as [tv'' [? ?]]; fold_Sub; fold erelGeVFun in *;
  tsteps; rewrite app_comm_cons; try rewrite (app_nil_end (compileE _ true)); 
  EApply;
  [ instantiate (1:= consMap tv'' te); 
    split; by [rewrite hdConsMap | rewrite tlConsMap]
  | revert H8; unfold_SubVE; rewrite H12; intro;
    destruct H8; destruct x; first (by inversion H8);
    eexists; eapply (Ev_step_rev H8); 
    eapply eq_ind; first (by constructor); simpl_Sub
  | red; intros; tsteps; apply H9; auto ] )
).

(******* 
   LET
 *******)
destruct_pairs; split;
red; intros; rename e into tc1, e0 into tc2; destruct H as [tcs1 [? ?]];
[ destruct H0 as [tcs2 [? ?]] | destruct H1 as [tcs2 [? ?]] ];
exists (fun i => LET (tcs1 i) (tcs2 i));
[ split; first (by apply (ObsLe_LET _ _ _ _ _ H H0))
| split; first (by apply (ObsLe_LET _ _ _ _ _ H H1)) ];
(
do 4 red; intros; simpl; rewrite app_ass;
revert H6; unfold_SubVE; intro; destruct H6;
destruct (Treduce_let_inv (Ev_to_Treduce H6)) as [x1 [tvv [? [? ?]]]];
eapply H3;
[ apply H5
| eexists; apply (Treduce_to_Ev H9)
| red; intros; tsteps;
  revert H10; simpl_Sub; intro H10;
  rewrite -> app_ass, app_comm_cons; eapply H4; 
  [ instantiate (1:= consMap tvv te);
    split; by [rewrite hdConsMap | rewrite tlConsMap]
  | eexists; apply (Treduce_to_Ev H10)
  | red; intros; tsteps; apply H7; auto]]
).

(******* 
   APP
 *******)
destruct_pairs; split;
(
red; intros; rename v into tf, v0 into tv; 
destruct H as [tfs [? ?]]; destruct H0 as [tvs [? ?]];
exists (fun i => APP (tfs i) (tvs i)); 
split; first (by apply (ObsLe_APP _ _ _ _ _ H H0));
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H1 p e (c',e',s',d') (erelGeEFun_len H3)) as [f [[n ?] ?]]));
(withcesd c' e' s' d'
 (destruct (H2 p e (c',e',s',d') (erelGeEFun_len H3)) as [v [[n' ?] ?]]));
eapply terminate_kstepreduce_rev; first (by simpl in H6; apply H6);
eapply terminate_kstepreduce_rev; first (by simpl in H8; apply H8)
).
(* attachret = false *)
change (App::c) with ([App]++c). change (v::f::s) with ([v;f]++s). eapply H7. 
- apply H3. 
- fold erelGeVFun. apply H9. apply H3.
- apply H4.
- fold erelGeVFun (app e e0). red; intros. apply H5; auto.
(* attachret = true *)
assert (Htv:= H9 _ _ H3). assert (Htf:= H7 _ _ H3 _ _ Htv _ H4 _ _ _ _ H5).
simpl in Htf; depdestr f; solve 
[ tsteps; eapply terminate_kstepreduce; last (by apply Htf); by ksteps
| do 2 red; intros; destruct (stepsfor_inv (H10 1)) as [? [? ?]]; 
  depdestr H11 ].

(******* 
   TAPP
 *******)
split;
(
red; intros; rename v into tf;
destruct H as [tfs [? ?]]; exists (fun i => TAPP (tfs i) ty');
split; first (by apply (ObsLe_TAPP _ _ _ ty' H));
do 4 red; intros; 
(withcesd c' e' s' d'
 (destruct (H0 p e (c',e',s',d') (erelGeEFun_len H1)) as [fv [[n ?] ?]]));
simpl; rewrite app_ass; simpl;
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H4; apply H4);
assert (C0 : (nil:Env 0, subTy (sr_sub srel) (subTy [|ty'|] ty)) =
  (nil:Env 0, subTy (sr_sub (sr_cons 
    (subTy (sr_sub srel) ty') (erelGeVFun srel) srel)) ty))
 by (rewrite -> sr_sub_cons, subTy_ss, sr_singsub; auto);
rewrite (cons_app fv); by eapply H5; 
[ apply H1
| unfold iff; Hhide_concl; revert H2; unfold_SubVE;
  Hrefl (tv : SVal PVal); Helimc C; unfold_SubVE; intros; Hshow_concl;
  instantiate (1:= << SVal,PVal # C0 >> tv); 
  Hsimpl; Helimc C1; intros; revert H2; Hunify
| unfold iff; fold erelGeVFun (app e e0); red; intros; simpl; tsteps;
  apply H3; do 2 red in H6 |- *;
  Hrewrite (erelGeVFun_sub ty [|ty'|] srel);
  Hrefl (tv : SVal PVal); Helimc C1; intros;
  revert H6; Hsimpl; rewrite subSR_singsub_srel; Hunify]
).

(******* 
   UNPK
 *******)
destruct_pairs; split;
red; intros; rename v into tv, e into tc; destruct H as [tvs [? ?]];
[ destruct H0 as [tcs [? ?]] | destruct H1 as [tcs [? ?]] ];
exists (fun i => UNPK (tvs i) (tcs i)); 
[ split; first (by apply (ObsLe_UNPK _ _ _ _ _ _ H H0))
| split; first (by apply (ObsLe_UNPK _ _ _ _ _ _ H H1)) ];
(
do 4 red; intros; simpl; rewrite ! app_ass;
(withcesd c' e' s' d'
 (destruct (H2 p e (c',e',s',d') (erelGeEFun_len H4)) as [v [[n ?] ?]]));
destruct (H8 _ _ H4) as [t' [tr [tvv [? ?]]]]; fold subTy erelGeVFun in *;
eapply terminate_kstepreduce_rev; simpl; first (by simpl in H7; apply H7);
tsteps; revert H5; unfold_SubVE; intro; rewrite H9 in H5;
destruct H5; destruct x; first (by inversion H5);
rewrite -> app_ass, app_comm_cons;
assert (C0 : (subEnv (sr_sub srel) env, nil:Env 0) = 
  (subEnv (sr_sub (sr_cons t' tr srel)) (subEnv shsub env), nil))
 by (rewrite -> subEnv_ss, sr_sub_cons, consmap_shsub; auto);
assert (C1 : (nil: Env 0, subTy (sr_sub srel) ty2) = 
  (nil, subTy (sr_sub (sr_cons t' tr srel)) (subTy shsub ty2)))
 by (rewrite -> sr_sub_cons, subTy_ss, consmap_shsub; auto);
by eapply H3;
[ instantiate (2:= sr_cons _ tr srel); instantiate (1:= consMap _ _); 
  split; first (by rewrite hdConsMap; Apply);
  rewrite tlConsMap;
  Hhide_evars; Hrefl (hev0 : SSub PSub); fold_Sub; 
  Hrewritec (erelGeEFun_sub env shsub (sr_cons t' tr srel)); Hshow_evars;
  rewrite (subSR_shsub_cons tr srel); 
  instantiate (1:= << SSub,PSub # C0 >> te); Hsimplc
| replace_hyp H5 (Ev_step_rev H5 (step_Unpk _ _ _ _));
  Hhide_concl; revert H5; simpl_Sub; fold_Sub; simpl_Sub;
  Hrewrite (subSub_ss [|t'|] shsub te); rewrite consmap_shsub;
  Hrewrite (subSub_idsub te); Hsimpl; Helimc C2;
  Hrewrite (subExp_ss (tcs p) [|t'|] (subL (sr_sub srel))); 
  rewrite -> consmap_subL, idsub_ss, <- (sr_sub_cons tr srel);
  Hrefl (tvv : SVal PVal); Helimc C2; Hrefl (tv0 : SVal PVal); Helimc C5; 
  intros; Hshow_concl;
  instantiate (1 := << SVal,PVal # C1 >> tv0); instantiate (1 := p); 
  exists x; revert H5; Hunify
| red; intros; tsteps;
  apply H6; do 2 red in H11 |- *;
  revert H11; Hrewrite (erelGeVFun_sub ty2 shsub (sr_cons t' tr srel)); 
  rewrite subSR_shsub_cons; Helimc C ]
).
Qed.

(*
   Compiler Correctness
*)

Theorem compiler_correctness : forall u E T (tv: Val u E T) (tc:Exp u E T),
  #c (compileV tv) ==cc== ^tv /\ #c (compileE tc false) ==cc== tc.
Proof.
split.
- split; 
  [ apply cv_implies_cc_le, (proj1 (compiler_correctness_le _))
  | apply cv_implies_cc_ge, (proj1 (compiler_correctness_ge _)) ].
- split; 
  [ eapply compiler_correctness_le 
  | eapply compiler_correctness_ge ].
Qed.

End CompilerCorrectFunctor.
