(*==========================================================================
  Polymorphically typed lambda-calculus with recursion, pairs, and arithmetic
  Part of the formalization for

   Realizability and Compositional Compiler Correctness 
      for a Polymorphic Language
   N Benton & C-K Hur
   April 2010

  ==========================================================================*)

Require Import List Program FunctionalExtensionality.
Require Import util.
Require Import ssreflect ssrbool Heq.

Set Implicit Arguments.

(*====== coqdoc printing directives ========================================*)
(** printing --> %\ensuremath{\mathrel{\texttt{->}}}% *)
(** printing ** %\ensuremath{\mathrel{\texttt{*}}}% *)
(* rem printing Int %{\textsf{Int}}% *)
(* rem printing Bool %{\textsf{Bool}}% *)
(* rem printing Arrow %{\textsf{Arrow}}% *)
(* rem printing Prod %{\textsf{Prod}}% *)
(* rem printing Ty %{\textsf{Ty}}% *)
(* rem printing Env %{\textsf{Env}}% *)

(** printing env %{\ensuremath{\Gamma}}% *)
(** printing env' %{\ensuremath{\Gamma'}}% *)
(** printing ty %{\ensuremath{\tau}}% *)
(** printing ty1 %\ensuremath{\tau_1}% *)
(** printing ty2 %\ensuremath{\tau_2}% *)
(** printing ty' %\ensuremath{\tau'}% *)

(** printing e %{\ensuremath{M}}% *)
(** printing e' %{\ensuremath{M'}}% *)
(** printing e1 %{\ensuremath{M_1}}% *)
(** printing e2 %{\ensuremath{M_2}}% *)
(** printing te %{\ensuremath{N}}% *)

(** printing v %{\ensuremath{V}}% *)
(** printing v1 %{\ensuremath{V_1}}% *)
(** printing v2 %{\ensuremath{V_2}}% *)
(** printing tv %{\ensuremath{W}}% *)
(** printing tv' %{\ensuremath{W'}}% *)
(** printing v' %{\ensuremath{V'}}% *)

(*
   Tactic
*)

Tactic Notation "RewritesWith" tactic(T) := 
  (intros; simpl; T;
   repeat (match goal with | [H : context[eq _ _] |- _] => rewrite -> H end); 
   T; auto).

Ltac Rewrites := RewritesWith idtac.

(*==========================================================================
  Types and contexts
  ==========================================================================*)

Inductive OperationNat : Set := PLUS | MULT | MINUS.
Inductive OperationBool : Set := EQ | LE | GT.

Inductive varT : nat -> Type :=
  | zVAR : forall u, varT (S u)
  | sVAR : forall u, varT u -> varT (S u).

Implicit Arguments zVAR [[u]].

Definition varT_empty (A:Type) (v:varT 0): A.
Proof. intros; inversion v. Defined.

(* Types in Type Variable Context *)
Inductive Ty (u: nat) : Type :=
  | TYVAR : varT u -> Ty u
  | UNIT : Ty u
  | INT : Ty u
  | ARROW : Ty u -> Ty u -> Ty u
  | PROD : Ty u -> Ty u -> Ty u
  | SUM  : Ty u -> Ty u -> Ty u
  | ALL : Ty (S u) -> Ty u
  | EXIST : Ty (S u) -> Ty u
.
Definition BOOL u := SUM (UNIT u) (UNIT u).

Implicit Arguments UNIT [ [u] ].
Implicit Arguments INT [ [u] ].
Implicit Arguments BOOL [ [u] ].

Infix " *** " := PROD (right associativity, at level 55).
Infix " +++ " := SUM (right associativity, at level 56).
Infix " --> " := ARROW  (right associativity, at level 57).

Definition mapT (P: nat -> Type) u w : Type := varT u -> P w.
Definition renT u w : Type := mapT varT u w.
Definition subT u w : Type := mapT Ty u w.

Definition idren u: renT u u := (fun v => v).
Definition idsub u: subT u u := (fun v => TYVAR v).
Implicit Arguments idren [[u]].
Implicit Arguments idsub [[u]].

Definition renL u w (r:renT u w) : renT (S u) (S w) := 
  (fun var =>
    match var in varT p return renT (Peano.pred p) w -> varT (S w) with
      | zVAR _ => fun _ => zVAR
      | sVAR _ var' => fun r => sVAR (r var')
    end r).

Fixpoint renTy u w (ren: renT u w) (ty: Ty u) : Ty w :=
  match ty with
    | TYVAR v => TYVAR (ren v)
    | UNIT => UNIT
    | INT => INT
    | ARROW ty1 ty2 => ARROW (renTy ren ty1) (renTy ren ty2)
    | PROD ty1 ty2 => PROD (renTy ren ty1) (renTy ren ty2)
    | SUM ty1 ty2 => SUM (renTy ren ty1) (renTy ren ty2)
    | ALL ty => ALL (renTy (renL ren) ty)
    | EXIST ty => EXIST (renTy (renL ren) ty)
  end.

Definition shren u : renT u (S u) := (@sVAR _).
Implicit Arguments shren [[u]].

Definition subL u w (s:subT u w) : subT (S u) (S w) :=  fun v => 
  match v in varT p return subT (Peano.pred p) w -> Ty (S w) with
    | zVAR _ => fun _ => TYVAR zVAR
    | sVAR _ v' => fun s => renTy shren (s v')
  end s.

Fixpoint subTy u w (sub: subT u w) (ty: Ty u) {struct ty} : Ty w :=
  match ty with
    | TYVAR v => sub v
    | UNIT => UNIT
    | INT => INT
    | ARROW ty1 ty2 => ARROW (subTy sub ty1) (subTy sub ty2)
    | PROD ty1 ty2 => PROD (subTy sub ty1) (subTy sub ty2)
    | SUM ty1 ty2 => SUM (subTy sub ty1) (subTy sub ty2)
    | ALL ty => ALL (subTy (subL sub) ty)
    | EXIST ty => EXIST (subTy (subL sub) ty)
  end.

Definition mcr P u v w (sub: mapT P v w) (ren : renT u v) : mapT P u w := 
  (fun var => sub (ren var)).
Definition rcs u v w (ren2: renT v w) (sub1: subT u v) : subT u w :=
  (fun var => renTy ren2 (sub1 var)).
Definition scs u v w (sub2: subT v w) (sub1: subT u v) : subT u w := 
  (fun var => subTy sub2 (sub1 var)).
Infix "@ss@" := scs (at level 55).

Definition rts u w (ren: renT u w) : subT u w :=  mcr idsub ren.

Definition shsub u : subT u (S u) := rts shren.
Implicit Arguments shsub [[u]].

Lemma map_extensional : forall P u w (r1 r2 : @mapT P u w), 
  (forall (var:varT u), r1 var = r2 var) -> r1 = r2.
Proof. intros. extension var. apply H. Qed.

Tactic Notation "mapext" ident(var) := apply map_extensional; intro var.
Ltac mapext_destr := 
  let var := fresh "var" in mapext var; depdestr var; auto.

Lemma rts_renL: forall u w (ren: renT u w), rts (renL ren) = subL (rts ren).
Proof. intros; mapext_destr. Qed.

Lemma renTy_def: forall u w (ren: renT u w), renTy ren = subTy (rts ren).
Proof. 
intros; extension ty; revert w ren.
induction ty; RewritesWith (try (f_equal; rewrite rts_renL)).
Qed.

Lemma subL_succ: forall u w (sub: subT u w) var,
  (subL sub) (sVAR var) = subTy shsub (sub var).
Proof. intros; unfold shsub; rewrite <- renTy_def; auto. Qed.

Lemma subL_shsub: forall u w (sub: subT u w), 
  (subL sub) @ss@ shsub = shsub @ss@ sub.
Proof.
intros; mapext var; unfold scs, compose, shsub; rewrite <- renTy_def; auto.
Qed.

Lemma renL_rr: forall u v w (ren2: renT v w) (ren1: renT u v),
  mcr (renL ren2) (renL ren1) = renL (mcr ren2 ren1).
Proof. intros; mapext_destr. Qed.

Lemma renTy_rr: forall u (ty:Ty u) v w (ren2: renT v w) (ren1: renT u v),
  renTy ren2 (renTy ren1 ty) = renTy (mcr ren2 ren1) ty.
Proof. induction ty; RewritesWith (try (f_equal; rewrite renL_rr)). Qed.

Lemma subL_rs: forall u v w (ren: renT v w) (sub: subT u v), 
  rcs (renL ren) (subL sub) = subL (rcs ren sub).
Proof.
intros; mapext_destr. 
RewritesWith (try (unfold rcs; simpl; rewrite !renTy_rr)).
Qed.

Lemma subTy_rs: forall u (ty:Ty u) v w (ren: renT v w) (sub: subT u v),
  renTy ren (subTy sub ty) = subTy (rcs ren sub) ty.
Proof. induction ty; RewritesWith (try (do 2 f_equal; rewrite subL_rs)). Qed.

Lemma subL_sr: forall u v w (sub: subT v w) (ren: renT u v),
  mcr (subL sub) (renL ren) = subL (mcr sub ren).
Proof. intros; mapext_destr. Qed.

Lemma subTy_sr: forall u (ty:Ty u) v w (sub: subT v w) (ren: renT u v),
  subTy sub (renTy ren ty) = subTy (mcr sub ren) ty.
Proof. 
induction ty; RewritesWith (try (do 2 f_equal; rewrite subL_sr)).
Qed. 

Lemma subL_ss: forall u v w (sub2: subT v w) (sub1: subT u v), 
  (subL sub2) @ss@ (subL sub1) = subL (sub2 @ss@ sub1).
Proof.
intros; mapext_destr; unfold scs; simpl; by rewrite -> subTy_rs, subTy_sr.
Qed.

Lemma subTy_ss: forall u v w (sub2: subT v w) (sub1: subT u v) (ty:Ty u), 
  subTy sub2 (subTy sub1 ty) = subTy (sub2 @ss@ sub1) ty.
Proof. 
intros until ty; revert v w sub2 sub1. 
induction ty; RewritesWith (try (do 2 f_equal; rewrite subL_ss)).
Qed.

Lemma subL_idsub: forall u, subL (@idsub u) = idsub.
Proof. intros; mapext_destr. Qed.

Lemma subTy_idsub: forall u (ty: Ty u), subTy idsub ty = ty.
Proof. intros; induction ty; RewritesWith (try rewrite subL_idsub). Qed.

Lemma sss_assoc: forall u v w z (sub3:subT w z) (sub2:subT v w) (sub1:subT u v),
  (sub3 @ss@ sub2) @ss@ sub1 = sub3 @ss@ (sub2 @ss@ sub1).
Proof. intros; mapext var; unfold scs; by rewrite subTy_ss. Qed.

Lemma idsub_ss: forall u w (sub:subT w u), idsub @ss@ sub = sub.
Proof. intros; mapext var; unfold scs; by rewrite subTy_idsub.
Qed.

Lemma ss_idsub: forall u w (sub:subT w u), sub @ss@ idsub = sub.
Proof. by intros; mapext var. Qed.

Definition hdmap P u w (m: mapT P (S u) w) : P w := m zVAR.

Definition tlmap P u w (m: mapT P (S u) w) : mapT P u w := 
  fun var => m (sVAR var).

Definition nilmap {P} {u} : mapT P 0 u.
Proof. red; intros; inversion H. Defined.

Definition consmap P u w (hd: P w) (tl : mapT P u w) : mapT P (S u) w :=
  (fun var =>
    match var in varT p return (mapT P (Peano.pred p) w) -> P w with
      | zVAR _ => fun _ => hd
      | sVAR _ var' => fun tl' => tl' var'
    end (tl: mapT P (Peano.pred (S u)) w)).

Notation "[| x , .. , y |]" := (consmap x .. (consmap y idsub) ..).

Lemma consmap_shsub : forall u w (sub: subT u w) val,
  (consmap val sub) @ss@ shsub = sub.
Proof. intros; mapext_destr. Qed.

Lemma consmap_subL : forall u v w (sub': subT v w) (sub: subT u v) val,
  (consmap val sub') @ss@ subL sub = consmap val (sub' @ss@ sub).
Proof.
intros; mapext_destr; unfold scs; simpl.
by rewrite -> renTy_def, subTy_ss, consmap_shsub.
Qed.

Lemma consmapInv : forall P u w (m:mapT P (S u) w),
  m = consmap (hdmap m) (tlmap m). 
Proof. intros. mapext_destr. Qed.

Lemma sub_singsub : forall u w (sub: subT u w) t,
  sub @ss@ [| t |] = [| subTy sub t |] @ss@ subL sub.
Proof. intros; rewrite -> consmap_subL, idsub_ss; mapext_destr; Rewrites. Qed.

(** Env **)

Definition Env u := list (Ty u).

Fixpoint subEnv u w (sub: subT u w) (env: Env u) : Env w := 
  (* map (subTy sub) env *)
  match env with
    | nil => nil
    | T::TS => subTy sub T :: subEnv sub TS
  end.

Lemma subEnv_length: forall u w (sub:subT u w) env,
  length (subEnv sub env) = length env. 
Proof. induction env; Rewrites. Qed.

Lemma subEnv_app: forall u w (sub: subT u w) (E E':Env u),
  subEnv sub (E++E') = (subEnv sub E)++(subEnv sub E').
Proof. by induction E; intros; [ | simpl; f_equal; apply IHE ]. Qed.

Lemma subEnv_ss: forall u v w (sub2: subT v w) (sub1: subT u v) env,
  subEnv sub2 (subEnv sub1 env) = subEnv (sub2 @ss@ sub1) env.
Proof. induction env; RewritesWith (try rewrite subTy_ss). Qed.

Lemma subEnv_idsub: forall u (env:Env u), subEnv idsub env = env.
Proof. induction env; RewritesWith (try rewrite subTy_idsub). Qed.

(** Tactic **)

Ltac fold_sub := 
  unfold subTy, subEnv in *; 
  repeat ( fold subL subTy subEnv  in * ).

Ltac simpl_sub := 
  simpl; fold_sub; rewrite ?sss_assoc;
  repeat ( (*fold_sub;*)
    rewrite -> ?renTy_def, ?rts_renL, ?subTy_ss, ?subEnv_ss, ?subL_ss; 

    try rewrite -> ?(sss_assoc _ (consmap _ _) shsub); 
    try rewrite <- ?(sss_assoc (consmap _ _) shsub); 
    rewrite -> ?consmap_shsub;

    try rewrite -> ?(sss_assoc _ (consmap _ _) (subL _)); 
    try rewrite <- ?(sss_assoc (consmap _ _) (subL _)); 
    rewrite -> ?consmap_subL;

    try rewrite -> ?(sss_assoc _ (subL _) shsub); 
    try rewrite <- ?(sss_assoc (subL _) shsub); 
    rewrite -> ?subL_shsub;

    rewrite -> ?sub_singsub;
    rewrite -> ?subTy_idsub, ?subEnv_idsub, ?idsub_ss, ?ss_idsub, ?subL_idsub;
    rewrite -> ?sss_assoc;
    auto
  ).

(*==========================================================================
  Typed terms in context
  ==========================================================================*)

(* Var *)

Inductive VarT u : Env u -> Ty u -> Type := 
| ZVAR : forall env ty, VarT (ty :: env) ty
| SVAR : forall env ty' ty, VarT env ty -> VarT (ty' :: env) ty.

Implicit Arguments ZVAR [ [u] [env] [ty] ].

Definition SVar u (x: Env u * Ty u) : Type :=
  match x with (env, ty) => VarT env ty end.
Notation   PVar := (_,_) (only parsing).
Implicit Arguments SVar [[u]].

Inductive Val u (env : Env u) : Ty u -> Type :=
| VAR   ty : VarT env ty -> Val env ty
| VUNIT : Val env UNIT
| VINT : nat -> Val env INT
| PAIR  ty1 ty2 : Val env ty1 -> Val env ty2 -> Val env (ty1 *** ty2)
| INL   ty1 ty2 : Val env ty1 -> Val env (ty1 +++ ty2)
| INR   ty1 ty2 : Val env ty2 -> Val env (ty1 +++ ty2)
| LAM   ty1 ty2 : Exp (ty1 :: env) ty2 -> Val env (ty1 --> ty2)  
| REC   ty1 ty2 : Exp (ty1 :: ty1 --> ty2 :: env) ty2 -> Val env (ty1 --> ty2)
| TLAM  ty      : Val (u:=S u) (subEnv shsub env) ty -> Val env (ALL ty)
| PACK  ty ty'  : Val env (subTy [| ty' |] ty) -> Val env (EXIST ty)

with Exp u (env : Env u) : Ty u -> Type :=
| VAL  ty : Val env ty -> Exp env ty
| OPN  : OperationNat -> Val env INT -> Val env INT -> Exp env INT
| OPB  : OperationBool -> Val env INT -> Val env INT -> Exp env BOOL
| FST  ty1 ty2 : Val env (ty1 *** ty2) -> Exp env ty1
| SND  ty1 ty2 : Val env (ty1 *** ty2) -> Exp env ty2
| CASE ty1 ty2 ty : 
  Val env (ty1 +++ ty2) -> Exp (ty1::env) ty -> Exp (ty2::env) ty -> Exp env ty
| LET  ty1 ty2 : Exp env ty1 -> Exp (ty1 :: env) ty2 -> Exp env ty2 
| APP  ty1 ty2 : Val env (ty1 --> ty2) -> Val env ty1 -> Exp env ty2
| TAPP ty      : Val env (ALL ty) -> forall ty' : Ty u, 
  Exp env (subTy [| ty' |] ty) 
| UNPK ty1 ty2 : 
  Val env (EXIST ty1) -> 
  Exp (u:=S u) (ty1 :: subEnv shsub env) (subTy shsub ty2) -> 
  Exp env ty2
.

Definition VBOOL u (env:Env u) (b:bool) : Val env (SUM UNIT UNIT) := 
  if b then (INL _ (VUNIT env)) else (INR _ (VUNIT env)).

Notation "^ v" := (VAL v) (at level 50).

Implicit Arguments VUNIT [u env].
Implicit Arguments VINT [u env].
Implicit Arguments VBOOL [u env].
Implicit Arguments TLAM [u env ty].
Implicit Arguments PACK [u env ty ty'].
Implicit Arguments UNPK [u env ty1 ty2].
Implicit Arguments Val [].
Implicit Arguments Exp [].

Definition SVal u (x: Env u * Ty u) : Type := 
  match x with (env, ty) => Val u env ty end.
Notation   PVal := (_,_) (only parsing).
Implicit Arguments SVal [[u]].

Definition SExp u (x: Env u * Ty u) : Type := 
  match x with (env, ty) => Exp u env ty end.
Notation   PExp := (_,_) (only parsing).
Implicit Arguments SExp [[u]].

Scheme Val_ind2 := Induction for Val Sort Prop
  with Exp_ind2 := Induction for Exp Sort Prop.

Combined Scheme ValExp_ind from Val_ind2, Exp_ind2.

Fixpoint subVar u w (sub: subT u w) (env: Env u) ty  (var: VarT env ty) 
  : VarT (subEnv sub env) (subTy sub ty) 
  :=
  match var with
    | ZVAR _ _ => ZVAR
    | SVAR _ _ _ v => SVAR _ (subVar sub v)
  end.

Lemma subVar_succ: forall u w (sub: subT u w) env ty (var: VarT env ty) ty',
  subVar sub (SVAR ty' var) = SVAR _ (subVar sub var).
Proof. by []. Qed.

Definition subValC1 : forall u w (sub : subT u w) (env: Env u) (ty: Ty (S u)),
  (subEnv (subL sub) (subEnv shsub env), subTy (subL sub) ty) =
  (subEnv shsub (subEnv sub env), subTy (subL sub) ty).
Proof. by intros; simpl_sub. Defined.
Implicit Arguments subValC1 [u w sub env ty].

Definition subValC2 : forall u w (sub: subT u w) env ty t,
  (subEnv sub env, subTy sub (subTy [| t |] ty)) =
  (subEnv sub env, subTy [| subTy sub t |] (subTy (subL sub) ty)).
Proof. by intros; simpl_sub. Defined.
Implicit Arguments subValC2 [u w sub env ty t].

Definition subExpC1 : forall u w (sub: subT u w) env ty ty',
  (subEnv sub env , subTy [| subTy sub ty' |] (subTy (subL sub) ty)) = 
  (subEnv sub env , subTy sub (subTy [| ty' |] ty)).
Proof. by intros; simpl_sub. Defined.
Implicit Arguments subExpC1 [u w sub env ty ty'].

Definition subExpC2 : forall u w (sub : subT u w) (env: Env u) ty ty',
  (subEnv (subL sub) (ty::subEnv shsub env),
   subTy (subL sub) (subTy shsub ty')) =
  (subTy (subL sub) ty::subEnv shsub (subEnv sub env), 
   subTy shsub (subTy sub ty')).
Proof. by intros; simpl_sub. Defined.
Implicit Arguments subExpC2 [u w sub env ty ty'].

Fixpoint subVal u w (sub: subT u w) env ty  (tv: Val u env ty) : 
  Val _ (subEnv sub env) (subTy sub ty) :=
  match tv with
    | VAR _ var => VAR (subVar sub var)
    | VUNIT => VUNIT
    | VINT i => VINT i
    | PAIR _ _ v1 v2 => PAIR (subVal sub v1) (subVal sub v2)
    | INL _ _ v1 => INL _ (subVal sub v1)
    | INR _ _ v2 => INR _ (subVal sub v2)
    | LAM _ _ e => LAM (subExp sub e)
    | REC _ _ e => REC (subExp sub e)
    | TLAM _ v => TLAM (<< SVal,PVal # subValC1 >> subVal (subL sub) v)
    | PACK _ t v => PACK (<< SVal,PVal # subValC2 >> subVal sub v)
  end

with subExp u w (sub: subT u w) env ty  (te: Exp u env ty) :
  Exp _ (subEnv sub env) (subTy sub ty) :=
  match te with
    | VAL _ v => ^ subVal sub v
    | OPN op v1 v2 => OPN op (subVal sub v1) (subVal sub v2)
    | OPB op v1 v2 => OPB op (subVal sub v1) (subVal sub v2)
    | FST _ _ v => FST (subVal sub v)
    | SND _ _ v => SND (subVal sub v)
    | CASE _ _ _ v e1 e2 => CASE (subVal sub v) (subExp sub e1) (subExp sub e2)
    | LET _ _ e1 e2 => LET (subExp sub e1) (subExp sub e2)
    | APP _ _ v1 v2 => APP (subVal sub v1) (subVal sub v2)
    | TAPP _ f t => << SExp,PExp # subExpC1 >> TAPP (subVal sub f) (subTy sub t)
    | UNPK _ _ v e => 
      UNPK (subVal sub v) (<< SExp,PExp # subExpC2 >> subExp (subL sub) e)
  end.

Lemma VarT_inv : 
  forall u w (sub:subT u w) env ty (var: VarT (subEnv sub env) ty),
  exists ty', exists var': VarT env ty',
    subTy sub ty' = ty /\ {{ SVar,PVar # subVar sub var' == var }}.
Proof.
intros. induction env. inversion var.
depdestr var. 
exists a. exists (ZVAR (env:=env) (ty:=a)). split; auto.

destruct (IHenv var) as [ty' [var' [? ?]]].
eexists ty'. exists (SVAR a var'). split; auto.
Hsubst. auto.
Qed.

(*==========================================================================
  Variable-domain maps. 
  By instantiating P with Var we get renamings.
  By instantiating P with Val we get substitutions.
  ==========================================================================*)

Section VariableDomainMap.

  Definition VDMap := forall u, (Env u) -> (Ty u) -> Type.

  Variable P : VDMap.
  Implicit Arguments P [].

  Record MapOps :=
  {
    vr : forall u env ty, VarT env ty -> P u env ty;   
    vl : forall u env ty, P u env ty -> Val u env ty;
    wk : forall u env ty' ty, P u env ty -> P u (ty' :: env) ty;
    sb : forall u w (sub: subT u w) env ty, 
      P u env ty -> P w (subEnv sub env) (subTy sub ty)
  }.

  Variable ops: MapOps.

  Definition MapT u E1 E2 : Type := forall t, VarT E1 t -> P u E2 t.

  Definition SMap u (x: Env u * Env u) : Type := 
    match x with (E1, E2) => MapT E1 E2 end.
  Notation   PMap := (_,_) (only parsing).
  Implicit Arguments SMap [[u]].

  Lemma Map_extensional : forall u (E:Env u) E' (r1 r2 : MapT E E'), 
    (forall t (var:VarT _ t), r1 _ var = r2 _ var) -> r1 = r2.
  Proof. intros. args2 extension t var. apply H. Qed.

  Tactic Notation "Mapext" ident(var) := apply Map_extensional; intros ? var.
  Ltac Mapext_destr := 
    let Var := fresh "Var" in Mapext Var; depdestr Var; auto.

  Definition hdMap u E E' t (m:MapT (t::E) E') : P u E' t := m _ (ZVAR (ty:=t)).

  Definition tlMap u E E' t (m:MapT (t::E) E') : @MapT u E E' := 
    (fun t' v => m _ (SVAR t v)).

  Definition nilMap {u} {E:Env u} : (MapT nil E).
  Proof. red; intros. inversion H. Defined.

  Lemma nilMap_eq: forall u E (map: @MapT u nil E), map = nilMap.
  Proof. intros; Mapext_destr. Qed.

  Definition hdEnv u (E:Env u) :Ty u := 
    match E with | nil => ALL (TYVAR zVAR) | hd ::_ => hd end.

  Definition consMap u E E' t (v:P u E' t) (m:MapT E E') : MapT (t::E) E' :=
    fun t' (var:VarT (t::E) t') => 
      match var in VarT p t' 
      return P _ E' (hdEnv p) -> MapT (tail p) E' -> P _ E' t' with
      | ZVAR _ _ => fun v m => v
      | SVAR _ _ _ var => fun v m => m _ var
      end v m.

  Lemma hdConsMap : forall u env env' ty (v : P u env' ty) (s : MapT env env'), 
    hdMap (consMap v s) = v. 
  Proof. auto. Qed.

  Lemma tlConsMap : forall u env env' ty (v : P u env' ty) (s : MapT env env'),
    tlMap (consMap v s) = s. 
  Proof. intros. Mapext_destr. Qed.

  Lemma consMapInv : forall u env env' ty (m:@MapT u (ty :: env) env'), 
    m = consMap (hdMap m) (tlMap m). 
  Proof. intros. Mapext_destr. Qed.

  Lemma consMap_succ: 
    forall u E E' T (v:P u E' T) (m:MapT E E') T' (var: VarT E T'), 
    (consMap v m) _ (SVAR _ var) = m _ var. 
  Proof. auto. Qed.

  Definition liftMap u env env' ty (m : MapT env env') : 
    MapT (ty :: env) (ty :: env') :=
    fun ty' (var: VarT (ty::env) ty') =>
      match var in VarT p ty' 
      return MapT (tail p) env' -> P u (hdEnv p::env') ty' with
        | ZVAR _ _ => fun _ => vr ops ZVAR
        | SVAR _ _ _ x => fun m => wk ops _ (m _ x)
      end m.

  Lemma liftMap_succ: 
    forall u (env:Env u) env' (m: MapT env env') ty ty' (var:VarT env ty'), 
    liftMap m (SVAR ty var) = wk ops _ (m _ var).
  Proof. auto. Qed.

  Definition substMap u w (sub: subT u w) env env' (m: MapT env env') : 
    MapT (subEnv sub env) (subEnv sub env').
  Proof.
  induction env; intros.
  - exact nilMap.
  - red; intros t var. depdestr var.
    - exact (sb ops sub (m _ ZVAR)).
    - exact (IHenv _ (fun ty var => m _ (SVAR _ var)) _ var).
  Defined.
  Implicit Arguments substMap [u w env env'].

  Lemma substMap_subVar: forall u w (sub: subT u w) env env' 
    (m: MapT env env') ty (var: VarT env ty),
    (substMap sub m) _ (subVar sub var) =  sb ops sub (m _ var).
  Proof.
  intros; induction env; first by inversion var.
  depdestr var; first by reflexivity.
  specialize (IHenv ((fun ty0 var0 => m _ (SVAR a var0)))); simpl in IHenv. 
  by rewrite <- IHenv.
  Qed.

  Lemma hdMap_substMap: forall u w (sub:subT u w) E E' t (m:MapT (t::E) E'), 
    hdMap (substMap sub m) = sb ops sub (hdMap m).
  Proof. auto. Qed.

  Lemma tlMap_substMap: forall u w (sub:subT u w) E E' t (m:MapT (t::E) E'),
    tlMap (substMap sub m) = substMap sub (tlMap m).
  Proof. intros. Mapext_destr. Qed.

  Fixpoint MapVal u env env' (m: MapT env env') ty (tv : Val u env ty) : 
    Val u env' ty :=
  match tv with
    | VAR _ v => vl ops (m _ v)
    | VUNIT => VUNIT
    | VINT i => VINT i
    | PAIR _ _ e1 e2 => PAIR (MapVal m e1) (MapVal m e2)
    | INL _ _ v1 => INL _ (MapVal m v1)
    | INR _ _ v2 => INR _ (MapVal m v2)
    | LAM _ _ e => LAM (MapExp (liftMap m) e)
    | REC _ _ e => REC (MapExp (liftMap (liftMap m)) e)
    | TLAM _ v => TLAM (MapVal (substMap shsub m) v)
    | PACK _ t v => PACK (MapVal m v)
  end

  with MapExp u env env' (m: MapT env env') ty (te : Exp u env ty) : 
    Exp u env' ty :=
   match te with
    | VAL _ v => ^ MapVal m v
    | OPN op e1 e2 => OPN op (MapVal m e1) (MapVal m e2)
    | OPB op e1 e2 => OPB op (MapVal m e1) (MapVal m e2)
    | FST _ _ e => FST (MapVal m e)
    | SND _ _ e => SND (MapVal m e)
    | CASE _ _ _ v e1 e2 => 
      CASE (MapVal m v) (MapExp (liftMap m) e1) (MapExp (liftMap m) e2)
    | LET _ _ e1 e2 => LET (MapExp m e1) (MapExp (liftMap m) e2)
    | APP _ _ e1 e2 => APP (MapVal m e1) (MapVal m e2)
    | TAPP _ v t => TAPP (MapVal m v) t
    | UNPK _ t v e => UNPK (MapVal m v) (MapExp (liftMap (substMap shsub m)) e)
  end.

  Definition idMap u (env: Env u) : MapT env env := (@vr ops u env).

  Definition dropMap u (env env': Env u) ty (m: MapT (ty::env) env') :
    MapT env env' :=
    (fun _ var => m _ (SVAR _ var)).

End VariableDomainMap.

Tactic Notation "Mapext" ident(var) := apply Map_extensional; intros ? var.
Ltac Mapext_destr := 
  let Var := fresh "Var" in Mapext Var; depdestr Var; auto.

Ltac solution_simpl := simpl; unfold solution_left, eq_rect_r, eq_rect; simpl.

Implicit Arguments substMap [P u w env env'].
Implicit Arguments nilMap [[P] [u] [E]].

(*==========================================================================
  Variable renamings: Map Var
  ==========================================================================*)

Definition RenMap := 
  (@Build_MapOps (@VarT) (fun _ _ _ v => v) VAR SVAR (subVar)).

Definition RenT := @MapT (@VarT).
Definition RenVal := @MapVal _ RenMap.
Definition RenExp := @MapExp _ RenMap.
Definition RenL := @liftMap _ RenMap. 
Implicit Arguments RenL [u env env' t].
Definition subRen := @substMap _ RenMap.
Implicit Arguments subRen [u w env env' t].
Definition dropRen := @dropMap (@VarT).
Definition idRen := @idMap _ RenMap.
Implicit Arguments idRen [[u] [env]].

Definition SRen := SMap (@VarT).
Notation   PRen := (_,_) (only parsing).
Implicit Arguments SRen [[u]].

(*==========================================================================
  Substitution
  ==========================================================================*)

Definition SubMap := @Build_MapOps (Val) VAR 
  (fun _ _ _ v => v) (fun u env ty' => RenVal ((SVAR ty'))) (subVal). 

Definition SubT := @MapT (Val).
Definition SubVal := @MapVal _ SubMap.
Definition SubExp := @MapExp _ SubMap.
Definition SubL := @liftMap _ SubMap.
Implicit Arguments SubL [u env env' t].
Definition subSub := @substMap _ SubMap.
Implicit Arguments subSub [u w env env' t].
Definition dropSub := @dropMap (Val).
Definition idSub := @idMap _ SubMap.
Implicit Arguments idSub [[u] [env]].

Definition SSub := SMap (Val).
Notation   PSub := (_,_) (only parsing).
Implicit Arguments SSub [[u]].

Definition MCR u (env:Env u) env' env'' P 
  (m:MapT P env' env'') (r:RenT env env') : MapT P env env'' :=
  (fun ty var => m _ (r _ var)).
Definition RCS u (env:Env u) env' env'' 
  (r:RenT env' env'') (s:SubT env env') : SubT env env'' := 
  (fun ty var => RenVal r (s _ var)).
Definition SCS u (env env' env'':Env u) 
  (s' : SubT env' env'') (s : SubT env env') : SubT env env'' := 
  (fun ty var => SubVal s' (s _ var)). 

Infix "@SS@" := SCS (at level 55, right associativity).

Definition RTS u (env:Env u) env' (R : RenT env env') : SubT env env' := 
  MCR idSub R.

Definition shRen u (env:Env u) ty : RenT env (ty::env) := @SVAR _ _ _.

Definition shSub u (env:Env u) ty : SubT env (ty::env) := RTS (@SVAR _ _ _).

Notation "'{|' x , .. , y '|}'" := (consMap x .. (consMap y idSub) ..).

Lemma RTS_idRen u (env:Env u): forall , RTS idRen = @idSub _ env. 
Proof. intros. Mapext_destr. Qed.

Lemma RTS_SS: 
  forall u (env env' env'':Env u) (ren:RenT env' env'') (ren':RenT env _), 
  RTS ren @SS@ RTS ren' = RTS (MCR ren ren'). 
Proof. auto. Qed.

Lemma fold_shSub: forall u E T, RTS (SVAR _) = @shSub u E T.
Proof. by []. Qed.

Lemma shSub_simpl: forall u (E: Env u) ty T (var:VarT E T), 
  (shSub ty) _ var = VAR (SVAR _ var). 
Proof. auto. Qed.

Lemma nilMap_idSub: forall u, @nilMap Val u [] = idSub.
Proof. by intros; Mapext_destr. Qed.

Ltac fold_Sub := fold_sub; rewrite -> ?fold_shSub;
  repeat (fold subVal subExp RenT RenL subRen dropRen (@idRen) 
               SubT SubL subSub dropSub (@idSub) in * ).

Ltac fold_fix Name unfoldcmd := 
  simpl;
  let temp := fresh "__temp__" in 
  pose (temp:= Name); unfoldcmd temp; simpl in temp;
  match goal with [ _ := ?body |- _ ] => change body with Name end;
  clear temp.

Ltac fold_subVE := 
  fold_fix subVal ltac:(fun X => unfold subVal in X);
  fold_fix subExp ltac:(fun X => unfold subExp in X).

Ltac fold_RenVE := 
  fold_fix RenVal ltac:(fun X => unfold RenVal, MapVal in X); 
  fold_fix RenExp ltac:(fun X => unfold RenExp, MapExp in X).

Ltac fold_SubVE :=
  fold_fix SubVal ltac:(fun X => unfold SubVal, MapVal in X);
  fold_fix SubExp ltac:(fun X => unfold SubExp, MapExp in X).

Ltac unfold_RenVE := 
  unfold RenExp, RenVal, MapExp, MapVal; fold_RenVE; fold_Sub.
Ltac unfold_SubVE :=
  unfold SubExp, SubVal, MapExp, MapVal; fold_SubVE; fold_Sub.

(*
   RenL subRen RenExp RenExp definitions
*)

Lemma RTS_RenL: forall u (env env' : Env u) (ty : Ty u) (R: RenT env env'),
  RTS (RenL _ R) = SubL ty (RTS R).
Proof. intros. Mapext_destr. Qed.

Lemma RTS_subRen: forall u w (sub:subT u w) (E E': Env u) (R: RenT E E'),
  RTS (subRen sub R) = subSub sub (RTS R).
Proof.
intros. Mapext var. destruct (VarT_inv _ _ var) as [t' [var' [? ?]]]. Hsubst.
induction var'; try apply  (IHvar' (dropRen R)); auto.
Qed.

Lemma RenValExp_def: forall u env,
  (forall env' (R: @RenT u env env'), RenVal R = SubVal (RTS R)) /\
  (forall env' (R: @RenT u env env'), RenExp R = SubExp (RTS R)).
Proof.
cut (forall u env, 
  (forall ty (val:Val u env ty) env' (R: RenT env env'),
    RenVal R val = SubVal (RTS R) val) /\
  (forall ty (exp:Exp   u env ty) env' (R: RenT env env'),
    RenExp R exp = SubExp (RTS R) exp)).
intros; split; repeat (intros; apply (@functional_extensionality_dep _));
 intros; apply (H _ _).

apply ValExp_ind; 
  intros; simpl; unfold RenVal, RenExp, MapVal, MapExp; 
  RewritesWith(rewrite ?RTS_RenL; rewrite ?RTS_subRen); auto.
Qed.
Definition RenVal_def u env := proj1 (@RenValExp_def u env).
Definition RenExp_def u env := proj2 (@RenValExp_def u env).

(*
   ss composition
*)

Lemma subVar_ss: 
  forall u env v w (sub2:subT v w) (sub1:subT u v) ty (var: VarT env ty),
  {{ SVar,PVar # subVar sub2 (subVar sub1 var) == subVar (sub2 @ss@ sub1) var }}.
Proof.
induction env.
- intros; inversion var.
- intros; depdestr var; Rewrites.
  - by fold_Sub; rewrite -> subEnv_ss, subTy_ss.
  - by Hrewritec (IHenv _ _ sub2 sub1 ty var); simpl_sub.
Qed.

Lemma subValExp_ss: forall u env, 
  (forall ty (val: Val u env ty) v w (sub2:subT v w) (sub1:subT u v),
   {{SVal,PVal # subVal sub2 (subVal sub1 val) == subVal (sub2 @ss@ sub1) val}})
  /\
  (forall ty (exp: Exp u env ty) v w (sub2:subT v w) (sub1:subT u v),
   {{SExp,PExp # subExp sub2 (subExp sub1 exp) == subExp (sub2 @ss@ sub1) exp}}).
Proof.
apply ValExp_ind; intros; simpl; fold_Sub;
  try Hrewritec (H _ _ sub2 sub1); 
  try Hrewritec (H0 _ _ sub2 sub1); 
  try Hrewritec (H1 _ _ sub2 sub1);
  simpl_sub.

- by Hrewritec (subVar_ss sub2 sub1 v).
- Hsimpl; Helimc C0; Hrewritec (H _ _ (subL sub2) (subL sub1)).
  by rewrite -> subL_ss, (subEnv_ss sub2 sub1); Hunify.
- Hsimpl; Helimc C0; Hrewritec (H _ _ sub2 sub1). 
  by rewrite -> (subTy_ss sub2 sub1), (subTy_ss (subL sub2) (subL sub1)),
   subL_ss; Hunify.
- Hsimplc; simpl; Hsimpl; fold_Sub; rewrite (subTy_ss sub2 sub1 ty').
  by Hrewritec (H _ _ sub2 sub1).
- rewrite (subTy_ss sub2 sub1); Helimc C.
  Hrewritec (H0 _ _ (subL sub2) (subL sub1)); Helimc C3. 
  by simpl_sub; Hunify.
Qed.
Definition subVal_ss u env := proj1 (@subValExp_ss u env).
Definition subExp_ss u env := proj2 (@subValExp_ss u env).

Lemma subRen_ss : forall (u : nat) env v w 
  (sub2:subT v w) (sub1:subT u v) env' (Ren : RenT env env'),
  {{ SRen,PRen # subRen sub2 (subRen sub1 Ren) == subRen (sub2 @ss@ sub1) Ren }}.
Proof.
induction env; intros.
- simpl_sub.
- rewrite (consMapInv (subRen sub2 (subRen sub1 Ren))).
  rewrite (consMapInv (subRen (sub2 @ss@ sub1) Ren)).
  fold_Sub; unfold subRen; rewrite -> ! hdMap_substMap, ! tlMap_substMap.
  simpl; rewrite -> (Hrw (subVar_ss _ _ _)), (Hrw (IHenv _ _ _ _ _ _)); Hsimplc.
  by Helimc.
Qed.

Lemma subSub_ss: forall u (env env': Env u) v w 
  (sub2: subT v w) (sub1: subT u v)  (S:SubT env env'),
  {{ SSub,PSub # subSub sub2 (subSub sub1 S) == subSub (sub2 @ss@ sub1) S }}.
Proof.
induction env; intros. 
- by simpl_sub.
- rewrite (consMapInv (subSub (sub2 @ss@ sub1) S)).
  rewrite (consMapInv (subSub sub2 (subSub sub1 S))).
  fold_Sub; unfold subSub; rewrite -> ! hdMap_substMap, ! tlMap_substMap.
  simpl; rewrite -> (Hrw (subVal_ss _ _ _)), (Hrw (IHenv _ _ _ _ _ _)); Hsimplc.
  by Helimc.
Qed.

Lemma consMap_ZVAR_shSub: forall u (E:Env u) T, 
  consMap (VAR (@ZVAR _ E _)) (shSub T) = idSub.
Proof. by intros; Mapext_destr. Qed.

Lemma RenL_succ: 
  forall u (env:Env u) env' (R: RenT env env') ty ty' (var:VarT env ty'), 
  RenL _ R (SVAR ty var) = SVAR _ (R _ var).
Proof. by apply liftMap_succ. Qed.

Lemma SubL_succ: 
  forall u (env:Env u) env' (S: SubT env env') ty ty' (var:VarT env ty'), 
  SubL _ S (SVAR ty var) = SubVal (shSub _) (S _ var).
Proof. by intros; unfold shSub; rewrite <- RenVal_def; apply liftMap_succ. Qed.

Lemma subSub_dropSub: forall u w (sub: subT u w) ty ty' env env' 
  (S: SubT (ty'::env) env') (var: VarT _ ty),
  subSub sub (dropSub S) var = subSub sub S (SVAR _ var).
Proof. by []. Qed.

Lemma hdMap_subSub : 
  forall u w (sub:subT u w) (E: Env u) E' t (m:SubT (t::E) E'), 
  hdMap (subSub sub m) = subVal sub (hdMap m).
Proof. by apply hdMap_substMap. Qed.

Lemma tlMap_subSub : 
  forall u w (sub:subT u w) (E: Env u) E' t (m:SubT (t::E) E'), 
  tlMap (subSub sub m) = subSub sub (tlMap m).
Proof. by apply tlMap_substMap. Qed.

Lemma subRen_subVar: forall u w (sub: subT u w) (env env': Env u) 
  (R: RenT env env') ty (var: VarT _ ty),
  subRen sub R (subVar sub var) = subVar sub (R _ var).
Proof. by apply substMap_subVar. Qed.

Lemma subSub_subVar: forall u w (sub: subT u w) (env env': Env u) 
  (S: SubT env env') ty (var:VarT _ ty),
  subSub sub S (subVar sub var) = subVal sub (S _ var).
Proof. by apply substMap_subVar. Qed.

Lemma subSub_shSub: forall u (env: Env u) w (sub: subT u w) ty,
  subSub sub (shSub (env:=env) ty) = shSub (subTy sub ty).
Proof. 
intros; Mapext var; destruct (VarT_inv _ _ var) as [ty' [var' [? ?]]].
by Hsubst; rewrite subSub_subVar.
Qed.

Lemma subRen_RenL: forall u env w (sub:subT u w) env' (Ren: RenT env env') ty,
  subRen sub (RenL ty Ren) = RenL _ (subRen sub Ren).
Proof.
intros; Mapext_destr; destruct (VarT_inv _ _ Var) as [ty' [var' [? ?]]]; Hsubst.
by rewrite -> RenL_succ, <- subVar_succ, subRen_subVar, 
 subRen_subVar, <- subVar_succ. 
Qed.

Lemma subValExp_RenValExp: forall u env,
  (forall ty (tm: Val u env ty) w (sub: subT u w) env' (Ren: RenT env env'),
    subVal sub (RenVal Ren tm) = RenVal (subRen sub Ren) (subVal sub tm))
  /\
  (forall ty (tm: Exp u env ty) w (sub: subT u w) env' (Ren: RenT env env'),
    subExp sub (RenExp Ren tm) = RenExp (subRen sub Ren) (subExp sub tm)).
Proof.
apply ValExp_ind; 
  intros; simpl; unfold_RenVE;
  RewritesWith (try rewrite subRen_subVar; try rewrite ! subRen_RenL); auto.

- Hrewritec (subRen_ss (subL sub) (shsub) Ren).
  Hrewritec (subRen_ss shsub sub Ren).
  by rewrite subL_shsub; Helimc C1; Helimc -> C2; Hunify.

- by Hsimpl; Hrefl (subVal sub v : SVal PVal) at 1; Helimc -> C1; Hunify.

- by apply (Heq_eq SExp PExp); Hsimplc.

- Hrewritec (subRen_ss (subL sub) shsub Ren). 
  Hrewritec (subRen_ss shsub sub Ren); rewrite subL_shsub.
  Hrefl (subExp (subL sub) e : SExp PExp) at 1; simpl; Helimc -> C3.
  repeat hrewrite (subEnv_ss (subL sub) shsub env); fold_Sub.
  by rewrite subL_shsub; Hunify.
Qed.
Definition subVal_RenVal u env := proj1 (@subValExp_RenValExp u env).
Definition subExp_RenExp u env := proj2 (@subValExp_RenValExp u env).

Lemma subSub_SubL: forall u env w (sub:subT u w) env' (Sub: SubT env env') ty,
  subSub sub (SubL ty Sub) = SubL _ (subSub sub Sub).
Proof.
intros; Mapext_destr; destruct (VarT_inv _ _ Var) as [ty' [var' [? ?]]]. 
Hsubst; rewrite -> SubL_succ, subSub_subVar, <- subVar_succ, 
 subSub_subVar, SubL_succ. 
by unfold shSub at 1; rewrite <- RenVal_def, subVal_RenVal, RenVal_def, 
 RTS_subRen, subSub_shSub.
Qed.

Lemma subValExp_SubValExp: forall u env,
  (forall ty (tm: Val u env ty) w (sub: subT u w) env' (Sub: SubT env env'),
    subVal sub (SubVal Sub tm) =  SubVal (subSub sub Sub) (subVal sub tm))
  /\
  (forall ty (tm: Exp u env ty) w (sub: subT u w) env' (Sub: SubT env env'),
    subExp sub (SubExp Sub tm) =  SubExp (subSub sub Sub) (subExp sub tm)).
Proof.
apply ValExp_ind;
  intros; simpl; unfold_SubVE;
  RewritesWith (try rewrite subSub_subVar; try rewrite ! subSub_SubL); auto.  

- Hrewritec (subSub_ss (subL sub) (shsub) Sub). 
  Hrewritec (subSub_ss shsub sub Sub).
  by rewrite subL_shsub; Helimc C1; Helimc -> C2; Hunify.
- by Hsimpl; Hrefl (subVal sub v : SVal PVal) at 1; Helimc -> C1; Hunify.
- by apply (Heq_eq SExp PExp); Hsimplc.
- Hrewritec (subSub_ss (subL sub) shsub Sub).
  Hrewritec (subSub_ss shsub sub Sub). rewrite subL_shsub.
  Hrefl (subExp (subL sub) e : SExp PExp) at 1; simpl; Helimc -> C3.
  repeat hrewrite (subEnv_ss (subL sub) shsub env); fold_Sub.
  by rewrite subL_shsub; Hunify.
Qed.
Definition subVal_SubVal u env := proj1 (@subValExp_SubValExp u env).
Definition subExp_SubExp u env := proj2 (@subValExp_SubValExp u env).


(*==========================================================================
  Composition of substitutions
  ==========================================================================*)

Section CompMR.

  Variable P: VDMap.
  Variable ops: MapOps P.
  
  Lemma liftMap_MR : 
    forall u (env env' env'':Env u) ty (m:MapT P env' env'') (r:RenT env env'),
    MCR (liftMap ops m) (RenL ty r) = liftMap ops (MCR m r).
  Proof. intros; Mapext_destr. Qed.

  Lemma substMap_MR: forall u w (sub: subT u w) (env env' env'': Env u) 
    (m: MapT P env' env'') (r: RenT env env'),
    MCR (substMap ops sub m) (subRen sub r) = substMap ops sub (MCR m r).
  Proof.
  intros; Mapext var; unfold MCR; induction env; first by inversion var.
  depdestr var; solution_simpl.
  - apply substMap_subVar. 
  - apply IHenv.
  Qed.

  Lemma MapValExp_MR : forall u env,
    (forall ty (tv : Val u env ty) env' env'' (m:MapT P env' env'') 
      (r : RenT env env'),
      MapVal ops m (RenVal r tv) = MapVal ops (MCR m r) tv)
    /\ 
    (forall ty (tc : Exp u env ty) env' env'' (m:MapT P env' env'') 
      (r : RenT env env'),
      MapExp ops m (RenExp r tc) = MapExp ops (MCR m r) tc).
  Proof.
    apply ValExp_ind; 
      intros; simpl; unfold RenVal, RenExp, MapVal, MapExp;
      RewritesWith(try rewrite ! liftMap_MR; try rewrite ! substMap_MR).
  Qed.
  Definition MapVal_MR u env := proj1 (@MapValExp_MR u env).
  Definition MapExp_MR u env := proj2 (@MapValExp_MR u env).

End CompMR.

Lemma SubL_RS : 
  forall u (env:Env u) env' env'' ty (R:RenT env' env'') (S:SubT env env'), 
  RCS (RenL _ R) (SubL _ S) = SubL ty (RCS R S).
Proof.
by intros; Mapext_destr; unfold RCS; simpl; rewrite -> ! (MapVal_MR RenMap).
Qed.

Lemma subSub_RS: forall u w (sub:subT u w) (env1 env2 env3:Env u) 
  (R: RenT env2 env3) (S: SubT env1 env2),
  RCS (subRen sub R) (subSub sub S) = subSub sub (RCS R S).
Proof.
intros; Mapext var; induction env1; first by inversion var.
depdestr var.
- by unfold RCS; solution_simpl; rewrite subVal_RenVal.
- replace_hyp IHenv1 (IHenv1 (dropSub S) var).
  unfold RCS at 1 in IHenv1; rewrite subSub_dropSub in IHenv1.
  by unfold RCS at 1; rewrite IHenv1.
Qed.

Lemma SubValExp_RS : forall u env,
   (forall t (v:Val u env t) env' env'' (r:RenT env' env'') (s : SubT env env'),
     RenVal r (SubVal s v) = SubVal (RCS r s) v)
/\ (forall t (e:Exp u env t) env' env'' (r:RenT env' env'') (s : SubT env env'),
     RenExp r (SubExp s e) = SubExp (RCS r s) e).
Proof.
  apply ValExp_ind; 
    intros; simpl; unfold RenVal, SubVal, RenExp, SubExp, MapVal, MapExp;
    RewritesWith(try rewrite ! SubL_RS; try rewrite <- ! subSub_RS).
Qed.
Definition SubVal_RS u env := proj1 (@SubValExp_RS u env).
Definition SubExp_RS u env := proj2 (@SubValExp_RS u env).

Lemma SubL_SS : 
  forall u (env env' env'':Env u) (s' : SubT env' env'') (s : SubT env env') ty, 
  SubL _ s' @SS@ SubL ty s = SubL _ (s' @SS@ s).
Proof.
intros; Mapext_destr; unfold SCS; simpl. 
by rewrite SubVal_RS; unfold SubVal at 1; rewrite MapVal_MR.
Qed.

Lemma subSub_SS: forall u w (sub:subT u w) (env1 env2 env3:Env u) 
  (S2: SubT env2 env3) (S1: SubT env1 env2),
  subSub sub S2 @SS@ subSub sub S1 = subSub sub (S2 @SS@ S1).
Proof.
intros; Mapext var; induction env1; first by inversion var.
depdestr var.
- by unfold SCS; solution_simpl; rewrite subVal_SubVal.
- replace_hyp IHenv1 (IHenv1 (dropSub S1) var).
  unfold SCS at 1 in IHenv1; rewrite subSub_dropSub in IHenv1.
  by unfold SCS at 1; rewrite IHenv1.
Qed.

Lemma SubValExp_SS : forall u env,
  (forall ty (v : Val u env ty) env' env'' 
   (s' : SubT env' env'') (s : SubT env env'),
   SubVal s' (SubVal s v) = SubVal (s' @SS@ s) v) /\
  (forall ty (e : Exp u env ty) env' env'' 
  (s' : SubT env' env'') (s : SubT env env'),
  SubExp s' (SubExp s e) = SubExp (s' @SS@ s) e).
Proof.
apply ValExp_ind;
  intros; simpl; unfold RenVal, SubVal, RenExp, SubExp, MapVal, MapExp;
  RewritesWith (try rewrite ! RTySub_SS; try rewrite ! SubL_SS; 
   try rewrite ! subSub_SS). 
Qed.
Definition SubVal_SS u env := proj1 (@SubValExp_SS u env).
Definition SubExp_SS u env := proj2 (@SubValExp_SS u env).

(* idSub Composition *)

Lemma SubL_idSub : forall u (env: Env u) ty, SubL ty (@idSub _ env) = idSub.
Proof. intros; Mapext_destr. Qed.

Lemma subSub_idSub : forall u w (sub: subT u w) (env: Env u),
  subSub sub (@idSub _ env) = idSub.
Proof. 
intros; Mapext var; destruct (VarT_inv _ _ var) as [ty' [var' [? ?]]].
by Hsubst; apply substMap_subVar.
Qed.

Lemma SubValExp_idSub : forall u env,
   (forall ty (v : Val u env ty), SubVal idSub v = v) /\ 
   (forall ty (e : Exp u env ty), SubExp idSub e = e).
Proof.
  apply ValExp_ind; 
    intros; simpl; unfold_SubVE;
    RewritesWith (try rewrite ! subSub_idSub; try rewrite ! SubL_idSub).
Qed.
Definition SubVal_idSub u env := proj1 (@SubValExp_idSub u env).
Definition SubExp_idSub u env := proj2 (@SubValExp_idSub u env).

Lemma idSub_SS : forall u (env env':Env u) (s : SubT env env'), 
  idSub @SS@ s = s.
Proof. intros; Mapext var; unfold SCS; apply SubVal_idSub. Qed.

Lemma SS_idSub : forall u (env env':Env u) (s:SubT env env'), 
  s @SS@ idSub = s.
Proof. intros; Mapext_destr. Qed.

(*
   Operations
*)

Lemma consMap_shSub : forall u env env' ty (v:Val u env' ty) (s:SubT env env'),
  (consMap v s) @SS@ shSub ty = s.
Proof. intros; Mapext_destr. Qed.

Lemma consMap_SubL : forall u env env' env'' ty (v:Val u env'' ty) 
  (s':SubT env' env'') (s:SubT env env'), 
  (consMap v s') @SS@ SubL _ s = consMap v (s' @SS@ s).
Proof.
intros; Mapext_destr; unfold SCS; simpl. 
by rewrite -> RenVal_def, SubVal_SS, consMap_shSub.
Qed.

Lemma subSub_consMap: 
  forall u w (sub:subT u w) E E' (S:SubT E E') T (tv:Val u E' T),
  subSub sub (consMap tv S) = consMap (subVal sub tv) (subSub sub S).
Proof. 
intros; Mapext_destr; solution_simpl. 
by repeat f_equal; Mapext_destr.
Qed.

Lemma SubL_shSub: forall u (env env':Env u) (S: SubT env env') ty,
  SubL _ S @SS@ shSub ty = shSub ty @SS@ S.
Proof. by intros; Mapext var; unfold SCS, shSub; rewrite <- RenVal_def. Qed.

Lemma Sub_singSub : forall u (env env':Env u) (sub: SubT env env') ty tm,
  sub @SS@ {| tm |} = {| SubVal sub tm |} @SS@ (SubL ty sub).
Proof. by intros; rewrite -> consMap_SubL, idSub_SS; Mapext_destr. Qed.

Lemma subVar_idsub: forall u (env:Env u) ty (var: VarT env ty),
  {{ SVar,PVar # subVar idsub var == var }}.
Proof.
intros; induction var; simpl.
- by fold subEnv; rewrite -> subEnv_idsub, subTy_idsub.
- by rewrite subTy_idsub; Hrewritec IHvar.
Qed.

Lemma subValExp_idsub: forall u env,
  (forall ty (tv: Val u env ty), {{ SVal,PVal # subVal idsub tv == tv }}) /\
  (forall ty (tc: Exp u env ty), {{ SExp,PExp # subExp idsub tc == tc }}).
Proof.
apply ValExp_ind; 
  intros; simpl; fold_Sub; Hsimpl; 
  try rewrite ! subEnv_ss; try rewrite ! subEnv_idsub; try rewrite ! subL_idsub;
  try Hrewritec H; try rewrite ! subL_idsub; try Hrewritec H0; try Hrewritec H1;
  try rewrite ! subTy_idsub; Hclear; auto.
- by Hrewritec (subVar_idsub v).
Qed.
Definition subVal_idsub u env := proj1 (@subValExp_idsub u env).
Definition subExp_idsub u env := proj2 (@subValExp_idsub u env).

Lemma subSub_idsub: forall u (env env':Env u) (Sub: SubT env env'),
  {{ SSub,PSub # subSub idsub Sub == Sub }}.
Proof.
induction env; intros.
- by rewrite -> nilMap_eq, (nilMap_eq (subSub idsub Sub)); simpl_sub.
- rewrite -> consMapInv, (consMapInv (subSub idsub Sub)).
  rewrite (hdMap_subSub idsub); Hrewrite (subVal_idsub (hdMap Sub)). 
  rewrite (tlMap_subSub idsub); Hrewrite (IHenv _ (tlMap Sub)).
  by simpl_sub; Hclear.
Qed.

(* Associativity *)

Lemma SSS_assoc : forall u (env env' env'' env''':Env u) 
  (S3 : SubT env'' env''') (S2: SubT env' env'') (S1 : SubT env env'), 
  (S3 @SS@ S2) @SS@ S1 = S3 @SS@ (S2 @SS@ S1).
Proof. by intros; Mapext var; unfold SCS; rewrite SubVal_SS. Qed.

(* simpl Tactic *)

Ltac simpl_Sub :=
  simpl; fold_Sub; rewrite ?SSS_assoc;
  repeat (
    rewrite -> ?RenVal_def, ?RenExp_def, ?RTS_idRen, ?RTS_RenL, ?RTS_subRen,
            ?subVal_SubVal, ?subExp_SubExp, ?SubVal_SS, ?SubExp_SS, ?SubL_SS;

    try rewrite -> ?(SSS_assoc _ (consMap _ _) (shSub _)); 
    try rewrite <- ?(SSS_assoc (consMap _ _) (shSub _)); 
    rewrite -> ?consMap_shSub;

    try rewrite -> ?(SSS_assoc _ (consMap _ _) (SubL _ _)); 
    try rewrite <- ?(SSS_assoc (consMap _ _) (SubL _ _)); 
    rewrite -> ?consMap_SubL;

    try rewrite -> ?(SSS_assoc _ (SubL _ _) (shSub _)); 
    try rewrite <- ?(SSS_assoc (SubL _ _) (shSub _)); 
    rewrite -> ?SubL_shSub;

    rewrite -> ?Sub_singSub, ?shSub_simpl, 
               ?subSub_shSub, ?subSub_subVar, ?subSub_SubL, ?subSub_consMap; 
    rewrite -> ?SubVal_idSub, ?SubExp_idSub, ?idSub_SS, ?SS_idSub,
               ?subSub_idSub, ?nilMap_idSub, ?SubL_idSub;
    rewrite -> ?SSS_assoc; 
    auto
  ).

(*==========================================================================
  Closed forms
  ==========================================================================*)
Definition CExp (ty:Ty 0) := Exp 0 nil ty.
Definition CVal (ty:Ty 0) := Val 0 nil ty.
Definition EVal u (E:Env u) := SubT E nil.

(* RECN *)

Definition BotF u env ty1 ty2 : Val u env (ty1-->ty2) :=
  REC (APP (VAR (SVAR _ ZVAR)) (VAR ZVAR)).

Lemma subBotF: forall u w (s:subT u w) (env:Env u) ty1 ty2,
  subVal s (BotF env ty1 ty2) = BotF (subEnv s env) (subTy s ty1) (subTy s ty2).
Proof. auto. Qed.

Lemma SubBotF : forall u (env1 env2: Env u) ty1 ty2  (S: SubT env1 env2),
  SubVal S (BotF env1 ty1 ty2) = BotF _ _ _.
Proof. auto. Qed.

Fixpoint RECN u env ty1 ty2 n (tc: Exp u (ty1 :: ty1 --> ty2 :: env) ty2) 
  : Val u env (ty1 --> ty2) :=
  match n with
    | 0 => BotF _ _ _
    | S n => LAM (SubExp (SubL _ {| RECN n tc |}) tc)
  end.

Lemma subRECN : forall u w (s:subT u w) env ty1 ty2 
  (tc : Exp u (ty1 :: ty1-->ty2 :: env) ty2) n,
  subVal s (RECN n tc) = RECN n (subExp s tc).
Proof.
induction n; by [ | simpl; rewrite <- IHn; simpl_Sub].
Qed.

Lemma SubRECN : forall u env1 env2 ty1 ty2  (S: SubT env1 env2) 
  (tc : Exp u (ty1 :: ty1-->ty2 :: env1) ty2) n,
  SubVal S (RECN n tc) = RECN n (SubExp (SubL _ (SubL _ S)) tc). 
Proof.
induction n; by [ | unfold_SubVE; rewrite <- IHn; simpl_Sub].
Qed.

(*
   idxvar, idxVar
*)

Fixpoint idxvar u (x:varT u)  : nat :=
  match x with 
    | zVAR _ => 0
    | sVAR _ x' => S (idxvar x')
  end.

Fixpoint idxVar u (E:Env u) T (x:VarT E T) : nat :=
  match x with 
    | ZVAR _ _ => 0
    | SVAR _ _ _ x' => S (idxVar x')
  end.

Lemma idxvar_inj: forall u (var var': varT u),
  idxvar var = idxvar var' -> var = var'.
Proof. 
induction u; intros; first by inversion var.
depdestr var; depdestr var'; inversion H; auto.
by f_equal; EApply.
Qed.

Lemma idxVar_inj: forall u (E:Env u) T (var var': VarT E T),
  idxVar var = idxVar var' -> var = var'.
Proof. 
induction E; intros; first by inversion var.
depdestr var; depdestr var'; inversion H; auto.
by f_equal; EApply.
Qed.

Lemma idxvar_range: forall u (var: varT u), idxvar var < u.
Proof. induction var; simpl; auto with arith. Qed.

Lemma idxVar_range: forall u (E:Env u) T (var: VarT E T),
  idxVar var < length E.
Proof. induction var; simpl; auto with arith. Qed.

Lemma idxVar_subVar: forall u (E:Env u) T (var: VarT E T) w (sub:subT u w),
  idxVar (subVar sub var) = idxVar var.
Proof. by induction var; intros; simpl; [ | rewrite IHvar]. Qed.

(*
   Macros 
*)

Fixpoint iTVar n i {struct i} : varT (S n) :=
  match i with
    | 0 => zVAR 
    | S i => match n with
               | 0 => zVAR
               | S n => sVAR (iTVar n i)
             end
  end.

Notation "#TV i" := (TYVAR (iTVar _ i)) (at level 55).

Fixpoint projEnv u (T:Ty u) (E: Env u) i {struct i} : Ty u :=
  match i with
    | 0 => T
    | S i => match E with
               | nil => T
               | T'::E' =>  projEnv T' E' i
             end
  end.

Fixpoint iVar u T E i {struct i} : @VarT u (T::E) (projEnv T E i) :=
  match i return VarT (T::E) (projEnv T E i) with
    | 0 => ZVAR
    | S i => match E return VarT (T::E) (projEnv T E (S i)) with
               | nil => ZVAR
               | T'::E' => SVAR _ (iVar T' E' i)
             end
  end.

Notation "#V i" := (VAR (iVar _ _ i)) (at level 55).

Definition UP u (T: Ty u) := subTy shsub T.

Definition upVal u E T (tv: Val u E T) := subVal shsub tv.
Definition upExp u E T (tc: Exp u E T) := subExp shsub tc.

Fixpoint UPVal u E E' T (tv: Val u E T) {struct E'} : Val _ (E'++E) T :=
  match E' with
    | nil => tv
    | T'::E'' => (SubVal (shSub _) (UPVal E'' tv))
  end.

Fixpoint UPExp u E E' T (tc: Exp u E T) {struct E'} : Exp _ (E'++E) T :=
  match E' with
    | nil => tc
    | T'::E'' => (SubExp (shSub _) (UPExp E'' tc))
  end.

Lemma subL_UP: forall u w (sub: subT u w) (T:Ty u),
  subTy (subL sub) (UP T) = UP (subTy sub T).
Proof. intros; unfold UP; simpl_sub. Qed.

Lemma subProjEnv: forall u (E:Env u) T i w (sub:subT u w),
   subTy sub (projEnv T E i) = projEnv (subTy sub T) (subEnv sub E) i.
Proof. by induction E; destruct i; intros; try apply IHE. Qed.

Lemma subiVar: forall u w (sub:subT u w) T E i,
  {{SVar,PVar # subVar sub (iVar T E i) == iVar (subTy sub T) (subEnv sub E) i}}.
Proof.
by intros; revert T i; induction E; destruct i; simpl; try Hrewritec (IHE a i).
Qed.
